import dataclasses
import json
import os
from functools import cached_property, wraps
from typing import List, Optional, Sequence, Set, Tuple

from ens import ENS
from eth_account import Account
from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress
from eth_utils import ValidationError
from hexbytes import HexBytes
from packaging import version as semantic_version
from prompt_toolkit import HTML, print_formatted_text
from web3 import Web3
from web3.contract import Contract
from web3.exceptions import BadFunctionCallOutput

from gnosis.eth import (
    EthereumClient,
    EthereumNetwork,
    EthereumNetworkNotSupported,
    TxSpeed,
)
from gnosis.eth.clients import EtherscanClient, EtherscanClientConfigurationProblem
from gnosis.eth.constants import NULL_ADDRESS, SENTINEL_ADDRESS
from gnosis.eth.contracts import (
    get_erc20_contract,
    get_erc721_contract,
    get_safe_V1_1_1_contract,
    get_sign_message_lib_contract,
)
from gnosis.eth.eip712 import eip712_encode
from gnosis.eth.utils import get_empty_tx_params
from gnosis.safe import InvalidInternalTx, Safe, SafeOperationEnum, SafeTx
from gnosis.safe.api import TransactionServiceApi
from gnosis.safe.multi_send import MultiSend, MultiSendOperation, MultiSendTx
from gnosis.safe.safe_deployments import safe_deployments

from safe_cli.ethereum_hd_wallet import get_account_from_words
from safe_cli.operators.exceptions import (
    AccountNotLoadedException,
    ExistingOwnerException,
    FallbackHandlerNotSupportedException,
    GuardNotSupportedException,
    HashAlreadyApproved,
    InvalidFallbackHandlerException,
    InvalidGuardException,
    InvalidMasterCopyException,
    InvalidMigrationContractException,
    InvalidNonceException,
    NonExistingOwnerException,
    NotEnoughEtherToSend,
    NotEnoughSignatures,
    SafeAlreadyUpdatedException,
    SafeVersionNotSupportedException,
    SameFallbackHandlerException,
    SameGuardException,
    SameMasterCopyException,
    SenderRequiredException,
    ThresholdLimitException,
    UpdateAddressesNotValid,
)
from safe_cli.safe_addresses import (
    get_default_fallback_handler_address,
    get_last_sign_message_lib_address,
    get_safe_contract_address,
    get_safe_l2_contract_address,
)
from safe_cli.utils import choose_option_from_list, get_erc_20_list, yes_or_no_question

from ..contracts import safe_to_l2_migration
from .hw_wallets.hw_wallet import HwWallet
from .hw_wallets.hw_wallet_manager import HwWalletType, get_hw_wallet_manager


@dataclasses.dataclass
class SafeCliInfo:
    address: str
    nonce: int
    threshold: int
    owners: List[str]
    master_copy: str
    modules: List[str]
    fallback_handler: str
    guard: str
    balance_ether: int
    version: str

    def __str__(self):
        return (
            f"safe-version={self.version} nonce={self.nonce} threshold={self.threshold} owners={self.owners} "
            f"master-copy={self.master_copy} fallback-hander={self.fallback_handler} "
            f"modules={self.modules} balance-ether={self.balance_ether:.4f}"
        )


def require_tx_service(f):
    @wraps(f)
    def decorated(self, *args, **kwargs):
        if not self.safe_tx_service:
            print_formatted_text(
                HTML(
                    f"<ansired>No tx service available for "
                    f"network={self.network.name}</ansired>"
                )
            )
            if self.etherscan:
                url = f"{self.etherscan.base_url}/address/{self.address}"
                print_formatted_text(HTML(f"<b>Try Etherscan instead</b> {url}"))
        else:
            return f(self, *args, **kwargs)

    return decorated


def require_default_sender(f):
    """
    Throws SenderRequiredException if not default sender configured
    """

    @wraps(f)
    def decorated(self, *args, **kwargs):
        if not self.default_sender and not self.hw_wallet_manager.sender:
            raise SenderRequiredException()
        else:
            return f(self, *args, **kwargs)

    return decorated


class SafeOperator:
    address: ChecksumAddress
    node_url: str
    ethereum_client: EthereumClient
    ens: ENS
    network: EthereumNetwork
    etherscan: Optional[EtherscanClient]
    safe_tx_service: Optional[TransactionServiceApi]
    safe: Safe
    safe_contract: Contract
    safe_contract_1_1_0: Contract
    accounts: Set[LocalAccount] = set()
    default_sender: Optional[LocalAccount]
    executed_transactions: List[str]
    _safe_cli_info: Optional[SafeCliInfo]
    require_all_signatures: bool

    def __init__(self, address: ChecksumAddress, node_url: str):
        self.address = address
        self.node_url = node_url
        self.ethereum_client = EthereumClient(self.node_url)
        self.ens = ENS.from_web3(self.ethereum_client.w3)
        self.network: EthereumNetwork = self.ethereum_client.get_network()
        try:
            self.etherscan = EtherscanClient(self.network)
        except EtherscanClientConfigurationProblem:
            self.etherscan = None

        try:
            self.safe_tx_service = TransactionServiceApi.from_ethereum_client(
                self.ethereum_client
            )
        except EthereumNetworkNotSupported:
            self.safe_tx_service = None

        self.safe = Safe(address, self.ethereum_client)
        self.safe_contract = self.safe.contract
        self.safe_contract_1_1_0 = get_safe_V1_1_1_contract(
            self.ethereum_client.w3, address=self.address
        )
        self.accounts: Set[LocalAccount] = set()
        self.default_sender: Optional[LocalAccount] = None
        self.executed_transactions: List[str] = []
        self._safe_cli_info: Optional[SafeCliInfo] = None  # Cache for SafeCliInfo
        self.require_all_signatures = (
            True  # Require all signatures to be present to send a tx
        )
        self.hw_wallet_manager = get_hw_wallet_manager()

    @cached_property
    def last_default_fallback_handler_address(self) -> ChecksumAddress:
        """
        :return: Address for last version of default fallback handler contract
        """
        return get_default_fallback_handler_address(self.ethereum_client)

    @cached_property
    def last_safe_contract_address(self) -> ChecksumAddress:
        """
        :return: Last version of the Safe Contract. Use events version for every network but mainnet
        """
        if self.network == EthereumNetwork.MAINNET:
            return get_safe_contract_address(self.ethereum_client)
        else:
            return get_safe_l2_contract_address(self.ethereum_client)

    @cached_property
    def ens_domain(self) -> Optional[str]:
        # FIXME After web3.py fixes the middleware copy
        if self.network == EthereumNetwork.MAINNET:
            return self.ens.name(self.address)

    @property
    def safe_cli_info(self) -> SafeCliInfo:
        if not self._safe_cli_info:
            self._safe_cli_info = self.refresh_safe_cli_info()
        return self._safe_cli_info

    def refresh_safe_cli_info(self) -> SafeCliInfo:
        self._safe_cli_info = self.get_safe_cli_info()
        return self._safe_cli_info

    def is_version_updated(self) -> bool:
        """
        :return: True if Safe Master Copy is updated, False otherwise
        """

        last_safe_contract_address = self.last_safe_contract_address
        if self.safe_cli_info.master_copy == last_safe_contract_address:
            return True
        else:  # Check versions, maybe safe-cli addresses were not updated
            try:
                safe_contract_version = Safe(
                    last_safe_contract_address, self.ethereum_client
                ).retrieve_version()
            except (
                BadFunctionCallOutput
            ):  # Safe master copy is not deployed or errored, maybe custom network
                return True  # We cannot say you are not updated ¯\_(ツ)_/¯

            return semantic_version.parse(
                self.safe_cli_info.version
            ) >= semantic_version.parse(safe_contract_version)

    def load_cli_owners_from_words(self, words: List[str]):
        if len(words) == 1:  # Reading seed from Environment Variable
            words = os.environ.get(words[0], default="").strip().split(" ")
        parsed_words = " ".join(words)
        try:
            for index in range(100):  # Try first accounts of seed phrase
                account = get_account_from_words(parsed_words, index=index)
                if account.address in self.safe_cli_info.owners:
                    self.load_cli_owners([account.key.hex()])
            if not index:
                print_formatted_text(
                    HTML(
                        "<ansired>Cannot generate any valid owner for this Safe</ansired>"
                    )
                )
        except ValidationError:
            print_formatted_text(
                HTML("<ansired>Cannot load owners from words</ansired>")
            )

    def load_cli_owners(self, keys: List[str]):
        for key in keys:
            try:
                account = Account.from_key(
                    os.environ.get(key, default=key)
                )  # Try to get key from `environ`
                self.accounts.add(account)
                balance = self.ethereum_client.get_balance(account.address)
                print_formatted_text(
                    HTML(
                        f"Loaded account <b>{account.address}</b> "
                        f'with balance={Web3.from_wei(balance, "ether")} ether'
                    )
                )
                if (
                    not self.default_sender
                    and not self.hw_wallet_manager.sender
                    and balance > 0
                ):
                    print_formatted_text(
                        HTML(
                            f"Set account <b>{account.address}</b> as default sender of txs"
                        )
                    )
                    self.default_sender = account
            except ValueError:
                print_formatted_text(HTML(f"<ansired>Cannot load key={key}</ansired>"))

    def load_hw_wallet(
        self,
        hw_wallet_type: HwWalletType,
        derivation_path: str,
        template_derivation_path: str,
    ):
        if not self.hw_wallet_manager.is_supported_hw_wallet(hw_wallet_type):
            return None
        if derivation_path is None:
            ledger_accounts = self.hw_wallet_manager.get_accounts(
                hw_wallet_type, template_derivation_path
            )
            if len(ledger_accounts) == 0:
                return None

            option = choose_option_from_list(
                "Select the owner address", ledger_accounts
            )
            if option is None:
                return None
            _, derivation_path = ledger_accounts[option]
        address = self.hw_wallet_manager.add_account(hw_wallet_type, derivation_path)
        balance = self.ethereum_client.get_balance(address)

        print_formatted_text(
            HTML(
                f"Loaded account <b>{address}</b> "
                f'with balance={Web3.from_wei(balance, "ether")} ether.'
            )
        )

        if (
            not self.default_sender
            and not self.hw_wallet_manager.sender
            and balance > 0
        ):
            self.hw_wallet_manager.set_sender(hw_wallet_type, derivation_path)
            print_formatted_text(HTML(f"HwDevice {address} added as sender"))
        else:
            print_formatted_text(HTML(f"HwDevice {address} wasn't added as sender"))

    def load_ledger_cli_owners(
        self, derivation_path: str = None, legacy_account: bool = False
    ):
        if legacy_account:
            self.load_hw_wallet(HwWalletType.LEDGER, derivation_path, "44'/60'/0'/{i}")
        else:
            self.load_hw_wallet(
                HwWalletType.LEDGER, derivation_path, "44'/60'/{i}'/0/0"
            )

    def load_trezor_cli_owners(
        self, derivation_path: str = None, legacy_account: bool = False
    ):
        if legacy_account:
            self.load_hw_wallet(HwWalletType.TREZOR, derivation_path, "44'/60'/0'/{i}")
        else:
            self.load_hw_wallet(
                HwWalletType.TREZOR, derivation_path, "44'/60'/0'/0/{i}"
            )

    def unload_cli_owners(self, owners: List[str]):
        accounts_to_remove: Set[Account] = set()
        for owner in owners:
            for account in self.accounts:
                if account.address == owner:
                    if self.default_sender and self.default_sender.address == owner:
                        self.default_sender = None
                    accounts_to_remove.add(account)
                    break
        self.accounts = self.accounts.difference(accounts_to_remove)
        # Check if there are ledger owners
        if self.hw_wallet_manager.wallets and len(accounts_to_remove) < len(owners):
            accounts_to_remove = (
                accounts_to_remove | self.hw_wallet_manager.delete_accounts(owners)
            )

        if accounts_to_remove:
            print_formatted_text(
                HTML("<ansigreen>Accounts have been deleted</ansigreen>")
            )
        else:
            print_formatted_text(HTML("<ansired>No account was deleted</ansired>"))

    def show_cli_owners(self):
        accounts = self.accounts | self.hw_wallet_manager.wallets
        if not accounts:
            print_formatted_text(HTML("<ansired>No accounts loaded</ansired>"))
        else:
            for account in accounts:
                print_formatted_text(
                    HTML(
                        f"<ansigreen><b>Account</b> {account.address} loaded</ansigreen>"
                    )
                )
            if self.default_sender:
                print_formatted_text(
                    HTML(
                        f"<ansigreen><b>Default sender:</b> {self.default_sender.address}"
                        f"</ansigreen>"
                    )
                )
            elif self.hw_wallet_manager.sender:
                print_formatted_text(
                    HTML(
                        f"<ansigreen><b>HwDevice sender:</b> {self.hw_wallet_manager.sender}"
                        f"</ansigreen>"
                    )
                )
            else:
                print_formatted_text(
                    HTML("<ansigreen>Not default sender set </ansigreen>")
                )

    def approve_hash(self, hash_to_approve: HexBytes, sender: str) -> bool:
        sender_account = [
            account for account in self.accounts if account.address == sender
        ]
        if not sender_account:
            raise AccountNotLoadedException(sender)
        elif sender not in self.safe_cli_info.owners:
            raise NonExistingOwnerException(sender)
        elif self.safe.retrieve_is_hash_approved(
            self.default_sender.address, hash_to_approve
        ):
            raise HashAlreadyApproved(hash_to_approve, self.default_sender.address)
        else:
            sender_account = sender_account[0]
            transaction_to_send = self.safe_contract.functions.approveHash(
                hash_to_approve
            ).build_transaction(
                {
                    "from": sender_account.address,
                    "nonce": self.ethereum_client.get_nonce_for_account(
                        sender_account.address
                    ),
                }
            )
            if self.ethereum_client.is_eip1559_supported():
                transaction_to_send = self.ethereum_client.set_eip1559_fees(
                    transaction_to_send
                )
            call_result = self.ethereum_client.w3.eth.call(transaction_to_send)
            if call_result:  # There's revert message
                return False
            else:
                signed_transaction = sender_account.sign_transaction(
                    transaction_to_send
                )
                tx_hash = self.ethereum_client.send_raw_transaction(
                    signed_transaction["rawTransaction"]
                )
                print_formatted_text(
                    HTML(
                        f"<ansigreen>Sent tx with tx-hash {tx_hash.hex()} from owner "
                        f"{self.default_sender.address}, waiting for receipt</ansigreen>"
                    )
                )
                if self.ethereum_client.get_transaction_receipt(tx_hash, timeout=120):
                    return True
                else:
                    print_formatted_text(
                        HTML(
                            f"<ansired>Tx with tx-hash {tx_hash.hex()} still not mined</ansired>"
                        )
                    )
                    return False

    def sign_message(
        self,
        eip191_message: Optional[str] = None,
        eip712_message_path: Optional[str] = None,
    ) -> bool:
        if eip712_message_path:
            try:
                message = json.load(open(eip712_message_path, "r"))
                message_bytes = b"".join(eip712_encode(message))
            except ValueError:
                raise ValueError
        else:
            message = eip191_message
            message_bytes = eip191_message.encode("UTF-8")

        safe_message_hash = self.safe.get_message_hash(message_bytes)

        sign_message_lib_address = get_last_sign_message_lib_address(
            self.ethereum_client
        )
        contract = get_sign_message_lib_contract(self.ethereum_client.w3, self.address)
        sign_message_data = HexBytes(
            contract.functions.signMessage(message_bytes).build_transaction(
                get_empty_tx_params(),
            )["data"]
        )
        print_formatted_text(HTML(f"Signing message: \n {message}"))
        if self.prepare_and_execute_safe_transaction(
            sign_message_lib_address,
            0,
            sign_message_data,
            operation=SafeOperationEnum.DELEGATE_CALL,
        ):
            print_formatted_text(
                HTML(f"Message was signed correctly: {safe_message_hash.hex()}")
            )

    def add_owner(self, new_owner: str, threshold: Optional[int] = None) -> bool:
        threshold = threshold if threshold is not None else self.safe_cli_info.threshold
        if new_owner in self.safe_cli_info.owners:
            raise ExistingOwnerException(new_owner)
        else:
            # TODO Allow to set threshold
            transaction = self.safe_contract.functions.addOwnerWithThreshold(
                new_owner, threshold
            ).build_transaction({"from": self.address, "gas": 0, "gasPrice": 0})
            if self.execute_safe_internal_transaction(transaction["data"]):
                self.safe_cli_info.owners = self.safe.retrieve_owners()
                self.safe_cli_info.threshold = threshold
                return True
            return False

    def remove_owner(self, owner_to_remove: str, threshold: Optional[int] = None):
        threshold = threshold if threshold is not None else self.safe_cli_info.threshold
        if owner_to_remove not in self.safe_cli_info.owners:
            raise NonExistingOwnerException(owner_to_remove)
        elif len(self.safe_cli_info.owners) == threshold:
            raise ThresholdLimitException()
        else:
            index_owner = self.safe_cli_info.owners.index(owner_to_remove)
            prev_owner = (
                self.safe_cli_info.owners[index_owner - 1]
                if index_owner
                else SENTINEL_ADDRESS
            )
            transaction = self.safe_contract.functions.removeOwner(
                prev_owner, owner_to_remove, threshold
            ).build_transaction({"from": self.address, "gas": 0, "gasPrice": 0})
            if self.execute_safe_internal_transaction(transaction["data"]):
                self.safe_cli_info.owners = self.safe.retrieve_owners()
                self.safe_cli_info.threshold = threshold
                return True
            return False

    def send_custom(
        self,
        to: str,
        value: int,
        data: bytes,
        safe_nonce: Optional[int] = None,
        delegate_call: bool = False,
    ) -> bool:
        if value > 0:
            safe_balance = self.ethereum_client.get_balance(self.address)
            if safe_balance < value:
                raise NotEnoughEtherToSend(safe_balance)
        operation = (
            SafeOperationEnum.DELEGATE_CALL if delegate_call else SafeOperationEnum.CALL
        )
        return self.prepare_and_execute_safe_transaction(
            to, value, data, operation, safe_nonce=safe_nonce
        )

    def send_ether(self, to: str, value: int, **kwargs) -> bool:
        return self.send_custom(to, value, b"", **kwargs)

    def send_erc20(self, to: str, token_address: str, amount: int, **kwargs) -> bool:
        transaction = (
            get_erc20_contract(self.ethereum_client.w3, token_address)
            .functions.transfer(to, amount)
            .build_transaction({"from": self.address, "gas": 0, "gasPrice": 0})
        )
        return self.send_custom(
            token_address, 0, HexBytes(transaction["data"]), **kwargs
        )

    def send_erc721(self, to: str, token_address: str, token_id: int, **kwargs) -> bool:
        transaction = (
            get_erc721_contract(self.ethereum_client.w3, token_address)
            .functions.transferFrom(self.address, to, token_id)
            .build_transaction({"from": self.address, "gas": 0, "gasPrice": 0})
        )
        return self.send_custom(token_address, 0, transaction["data"], **kwargs)

    def change_fallback_handler(self, new_fallback_handler: str) -> bool:
        if new_fallback_handler == self.safe_cli_info.fallback_handler:
            raise SameFallbackHandlerException(new_fallback_handler)
        elif semantic_version.parse(
            self.safe_cli_info.version
        ) < semantic_version.parse("1.1.0"):
            raise FallbackHandlerNotSupportedException()
        elif (
            new_fallback_handler != NULL_ADDRESS
            and not self.ethereum_client.is_contract(new_fallback_handler)
        ):
            raise InvalidFallbackHandlerException(
                f"{new_fallback_handler} address is not a contract"
            )
        else:
            transaction = self.safe_contract.functions.setFallbackHandler(
                new_fallback_handler
            ).build_transaction({"from": self.address, "gas": 0, "gasPrice": 0})
            if self.execute_safe_internal_transaction(transaction["data"]):
                self.safe_cli_info.fallback_handler = new_fallback_handler
                self.safe_cli_info.version = self.safe.retrieve_version()
                return True

    def change_guard(self, guard: str) -> bool:
        if guard == self.safe_cli_info.guard:
            raise SameGuardException(guard)
        elif semantic_version.parse(
            self.safe_cli_info.version
        ) < semantic_version.parse("1.3.0"):
            raise GuardNotSupportedException()
        elif guard != NULL_ADDRESS and not self.ethereum_client.is_contract(guard):
            raise InvalidGuardException(f"{guard} address is not a contract")
        else:
            transaction = self.safe_contract.functions.setGuard(
                guard
            ).build_transaction({"from": self.address, "gas": 0, "gasPrice": 0})
            if self.execute_safe_internal_transaction(transaction["data"]):
                self.safe_cli_info.guard = guard
                self.safe_cli_info.version = self.safe.retrieve_version()
                return True

    def change_master_copy(self, new_master_copy: str) -> bool:
        if new_master_copy == self.safe_cli_info.master_copy:
            raise SameMasterCopyException(new_master_copy)
        else:
            safe_version = self.safe.retrieve_version()
            if semantic_version.parse(safe_version) >= semantic_version.parse("1.3.0"):
                raise SafeVersionNotSupportedException(
                    f"{safe_version} cannot be updated (yet)"
                )

            try:
                Safe(new_master_copy, self.ethereum_client).retrieve_version()
            except BadFunctionCallOutput:
                raise InvalidMasterCopyException(new_master_copy)

            transaction = self.safe_contract_1_1_0.functions.changeMasterCopy(
                new_master_copy
            ).build_transaction({"from": self.address, "gas": 0, "gasPrice": 0})
            if self.execute_safe_internal_transaction(transaction["data"]):
                self.safe_cli_info.master_copy = new_master_copy
                self.safe_cli_info.version = self.safe.retrieve_version()
                return True

    def update_version(self) -> Optional[bool]:
        """
        Update Safe Master Copy and Fallback handler to the last version

        :return:
        """

        safe_version = self.safe.retrieve_version()
        if semantic_version.parse(safe_version) >= semantic_version.parse("1.3.0"):
            raise SafeVersionNotSupportedException(
                f"{safe_version} cannot be updated (yet)"
            )

        if self.is_version_updated():
            raise SafeAlreadyUpdatedException(f"{safe_version} already updated")

        addresses = (
            self.last_safe_contract_address,
            self.last_default_fallback_handler_address,
        )
        if not all(
            self.ethereum_client.is_contract(contract) for contract in addresses
        ):
            raise UpdateAddressesNotValid(
                "Not valid addresses to update Safe", *addresses
            )

        multisend = MultiSend(ethereum_client=self.ethereum_client)
        tx_params = {"from": self.address, "gas": 0, "gasPrice": 0}
        multisend_txs = [
            MultiSendTx(MultiSendOperation.CALL, self.address, 0, data)
            for data in (
                self.safe_contract_1_1_0.functions.changeMasterCopy(
                    self.last_safe_contract_address
                ).build_transaction(tx_params)["data"],
                self.safe_contract_1_1_0.functions.setFallbackHandler(
                    self.last_default_fallback_handler_address
                ).build_transaction(tx_params)["data"],
            )
        ]

        multisend_data = multisend.build_tx_data(multisend_txs)

        if self.prepare_and_execute_safe_transaction(
            multisend.address,
            0,
            multisend_data,
            operation=SafeOperationEnum.DELEGATE_CALL,
        ):
            self.safe_cli_info.master_copy = self.last_safe_contract_address
            self.safe_cli_info.fallback_handler = (
                self.last_default_fallback_handler_address
            )
            self.safe_cli_info.version = self.safe.retrieve_version()

    def update_version_to_l2(
        self, migration_contract_address: ChecksumAddress
    ) -> Optional[bool]:
        """
        Update not L2 Safe to L2, so official UI supports it. Useful when replaying Safes deployed in
        non L2 networks (like mainnet) in L2 networks.
        Only v1.1.1, v1.3.0 and v1.4.1 versions are supported. Also, Safe nonce must be 0.

        :return:
        """

        if not self.ethereum_client.is_contract(migration_contract_address):
            raise InvalidMigrationContractException(
                f"Non L2 to L2 migration contract {migration_contract_address} is not deployed"
            )

        safe_version = self.safe.retrieve_version()
        chain_id = self.ethereum_client.get_chain_id()

        if self.safe.retrieve_nonce() > 0:
            raise InvalidNonceException("Nonce must be 0 for non L2 to L2 migration")

        l2_migration_contract = self.ethereum_client.w3.eth.contract(
            NULL_ADDRESS, abi=safe_to_l2_migration["abi"]
        )
        if safe_version == "1.1.1":
            safe_l2_singleton = safe_deployments["1.3.0"]["GnosisSafeL2"][str(chain_id)]
            fallback_handler = safe_deployments["1.3.0"][
                "CompatibilityFallbackHandler"
            ][str(chain_id)]
            data = HexBytes(
                l2_migration_contract.functions.migrateFromV111(
                    safe_l2_singleton, fallback_handler
                ).build_transaction(get_empty_tx_params())["data"]
            )
        elif safe_version in ("1.3.0", "1.4.1"):
            safe_l2_singleton = safe_deployments[safe_version]["GnosisSafeL2"][
                str(chain_id)
            ]
            fallback_handler = self.safe_cli_info.fallback_handler
            data = HexBytes(
                l2_migration_contract.functions.migrateToL2(
                    safe_l2_singleton
                ).build_transaction(get_empty_tx_params())["data"]
            )
        else:
            raise InvalidMasterCopyException(
                "Current version is not supported to migrate to L2"
            )

        if self.prepare_and_execute_safe_transaction(
            migration_contract_address,
            0,
            data,
            operation=SafeOperationEnum.DELEGATE_CALL,
        ):
            self.safe_cli_info.master_copy = safe_l2_singleton
            self.safe_cli_info.fallback_handler = fallback_handler
            self.safe_cli_info.version = self.safe.retrieve_version()

    def change_threshold(self, threshold: int):
        if threshold == self.safe_cli_info.threshold:
            print_formatted_text(
                HTML(f"<ansired>Threshold is already {threshold}</ansired>")
            )
        elif threshold > len(self.safe_cli_info.owners):
            print_formatted_text(
                HTML(
                    f"<ansired>Threshold={threshold} bigger than number "
                    f"of owners={len(self.safe_cli_info.owners)}</ansired>"
                )
            )
        else:
            transaction = self.safe_contract.functions.changeThreshold(
                threshold
            ).build_transaction({"from": self.address, "gas": 0, "gasPrice": 0})

            if self.execute_safe_internal_transaction(transaction["data"]):
                self.safe_cli_info.threshold = threshold

    def enable_module(self, module_address: str):
        if module_address in self.safe_cli_info.modules:
            print_formatted_text(
                HTML(f"<ansired>Module {module_address} is already enabled</ansired>")
            )
        else:
            transaction = self.safe_contract.functions.enableModule(
                module_address
            ).build_transaction({"from": self.address, "gas": 0, "gasPrice": 0})
            if self.execute_safe_internal_transaction(transaction["data"]):
                self.safe_cli_info.modules = self.safe.retrieve_modules()

    def disable_module(self, module_address: str):
        if module_address not in self.safe_cli_info.modules:
            print_formatted_text(
                HTML(f"<ansired>Module {module_address} is not enabled</ansired>")
            )
        else:
            pos = self.safe_cli_info.modules.index(module_address)
            if pos == 0:
                previous_address = SENTINEL_ADDRESS
            else:
                previous_address = self.safe_cli_info.modules[pos - 1]
            transaction = self.safe_contract.functions.disableModule(
                previous_address, module_address
            ).build_transaction({"from": self.address, "gas": 0, "gasPrice": 0})
            if self.execute_safe_internal_transaction(transaction["data"]):
                self.safe_cli_info.modules = self.safe.retrieve_modules()

    def print_info(self):
        for key, value in dataclasses.asdict(self.safe_cli_info).items():
            print_formatted_text(
                HTML(
                    f"<b><ansigreen>{key.capitalize()}</ansigreen></b>="
                    f"<ansiblue>{value}</ansiblue>"
                )
            )
        if self.ens_domain:
            print_formatted_text(
                HTML(
                    f"<b><ansigreen>Ens domain</ansigreen></b>="
                    f"<ansiblue>{self.ens_domain}</ansiblue>"
                )
            )
        if self.safe_tx_service:
            url = f"{self.safe_tx_service.base_url}/api/v1/safes/{self.address}/transactions/"
            print_formatted_text(
                HTML(
                    f"<b><ansigreen>Safe Tx Service</ansigreen></b>="
                    f"<ansiblue>{url}</ansiblue>"
                )
            )

        if self.etherscan:
            url = f"{self.etherscan.base_url}/address/{self.address}"
            print_formatted_text(
                HTML(
                    f"<b><ansigreen>Etherscan</ansigreen></b>="
                    f"<ansiblue>{url}</ansiblue>"
                )
            )

        if not self.hw_wallet_manager.is_supported_hw_wallet(HwWalletType.LEDGER):
            print_formatted_text(
                HTML(
                    "<b><ansigreen>Ledger</ansigreen></b>="
                    "<ansired>Disabled </ansired> <b>Optional ledger library is not installed, run pip install safe-cli[ledger] </b>"
                )
            )
        else:
            print_formatted_text(
                HTML(
                    "<b><ansigreen>Ledger</ansigreen></b>="
                    "<ansiblue>supported</ansiblue>"
                )
            )

        if not self.hw_wallet_manager.is_supported_hw_wallet(HwWalletType.TREZOR):
            print_formatted_text(
                HTML(
                    "<b><ansigreen>Trezor</ansigreen></b>="
                    "<ansired>Disabled </ansired> <b>Optional trezor library is not installed, run pip install safe-cli[trezor] </b>"
                )
            )
        else:
            print_formatted_text(
                HTML(
                    "<b><ansigreen>Trezor</ansigreen></b>="
                    "<ansiblue>supported</ansiblue>"
                )
            )

        if not self.is_version_updated():
            print_formatted_text(
                HTML(
                    "<ansired>Safe is not updated! You can use <b>update</b> command to update "
                    "the Safe to a newest version</ansired>"
                )
            )

    def get_safe_cli_info(self) -> SafeCliInfo:
        safe = self.safe
        balance_ether = Web3.from_wei(
            self.ethereum_client.get_balance(self.address), "ether"
        )
        safe_info = safe.retrieve_all_info()
        return SafeCliInfo(
            self.address,
            safe_info.nonce,
            safe_info.threshold,
            safe_info.owners,
            safe_info.master_copy,
            safe_info.modules,
            safe_info.fallback_handler,
            safe_info.guard,
            balance_ether,
            safe_info.version,
        )

    def get_threshold(self):
        print_formatted_text(self.safe.retrieve_threshold())

    def get_nonce(self):
        print_formatted_text(self.safe.retrieve_nonce())

    def get_owners(self):
        print_formatted_text(self.safe.retrieve_owners())

    def execute_safe_internal_transaction(self, data: bytes) -> bool:
        return self.prepare_and_execute_safe_transaction(self.address, 0, data)

    def prepare_safe_transaction(
        self,
        to: str,
        value: int,
        data: bytes,
        operation: SafeOperationEnum = SafeOperationEnum.CALL,
        safe_nonce: Optional[int] = None,
    ) -> SafeTx:
        safe_tx = self.safe.build_multisig_tx(
            to, value, data, operation=operation.value, safe_nonce=safe_nonce
        )
        self.sign_transaction(safe_tx)  # Raises exception if it cannot be signed
        return safe_tx

    def prepare_and_execute_safe_transaction(
        self,
        to: str,
        value: int,
        data: bytes,
        operation: SafeOperationEnum = SafeOperationEnum.CALL,
        safe_nonce: Optional[int] = None,
    ) -> bool:
        safe_tx = self.prepare_safe_transaction(
            to, value, data, operation, safe_nonce=safe_nonce
        )
        return self.execute_safe_transaction(safe_tx)

    @require_default_sender  # Throws Exception if default sender not found
    def execute_safe_transaction(self, safe_tx: SafeTx):
        try:
            if self.default_sender:
                call_result = safe_tx.call(self.default_sender.address)
            else:
                call_result = safe_tx.call(self.hw_wallet_manager.sender.address)
            print_formatted_text(HTML(f"Result: <ansigreen>{call_result}</ansigreen>"))
            if yes_or_no_question("Do you want to execute tx " + str(safe_tx)):
                if self.default_sender:
                    tx_hash, tx = safe_tx.execute(
                        self.default_sender.key, eip1559_speed=TxSpeed.NORMAL
                    )
                else:
                    tx_hash, tx = self.hw_wallet_manager.execute_safe_tx(
                        safe_tx, eip1559_speed=TxSpeed.NORMAL
                    )
                self.executed_transactions.append(tx_hash.hex())
                print_formatted_text(
                    HTML(
                        f"<ansigreen>Sent tx with tx-hash {tx_hash.hex()} "
                        f"and safe-nonce {safe_tx.safe_nonce}, waiting for receipt</ansigreen>"
                    )
                )
                tx_receipt = self.ethereum_client.get_transaction_receipt(
                    tx_hash, timeout=120
                )
                if tx_receipt:
                    fees = self.ethereum_client.w3.from_wei(
                        tx_receipt["gasUsed"]
                        * tx_receipt.get("effectiveGasPrice", tx.get("gasPrice", 0)),
                        "ether",
                    )
                    print_formatted_text(
                        HTML(
                            f"<ansigreen>Tx was executed on block-number={tx_receipt['blockNumber']}, fees "
                            f"deducted={fees}</ansigreen>"
                        )
                    )
                    self.safe_cli_info.nonce += 1
                    return True
                else:
                    print_formatted_text(
                        HTML(
                            f"<ansired>Tx with tx-hash {tx_hash.hex()} still not mined</ansired>"
                        )
                    )
        except InvalidInternalTx as invalid_internal_tx:
            print_formatted_text(
                HTML(f"Result: <ansired>InvalidTx - {invalid_internal_tx}</ansired>")
            )
        return False

    # Batch_transactions multisend
    def batch_safe_txs(
        self, safe_nonce: int, safe_txs: Sequence[SafeTx]
    ) -> Optional[SafeTx]:
        """
        Submit signatures to the tx service. It's recommended to be on Safe v1.3.0 to prevent issues
        with `safeTxGas` and gas estimation.

        :return:
        """

        try:
            multisend = MultiSend(ethereum_client=self.ethereum_client)
        except ValueError:
            multisend = None
            print_formatted_text(
                HTML(
                    "<ansired>Multisend contract is not deployed on this network and it's required for "
                    "batching txs</ansired>"
                )
            )

        multisend_txs = []
        for safe_tx in safe_txs:
            # Check if call is already a Multisend call
            inner_txs = MultiSend.from_transaction_data(safe_tx.data)
            if inner_txs:
                multisend_txs.extend(inner_txs)
            else:
                multisend_txs.append(
                    MultiSendTx(
                        MultiSendOperation.CALL, safe_tx.to, safe_tx.value, safe_tx.data
                    )
                )

        if len(multisend_txs) == 1:
            safe_tx.safe_tx_gas = 0
            safe_tx.base_gas = 0
            safe_tx.gas_price = 0
            safe_tx.signatures = b""
            safe_tx.safe_nonce = safe_nonce  # Resend single transaction
        elif multisend:
            safe_tx = SafeTx(
                self.ethereum_client,
                self.address,
                multisend.address,
                0,
                multisend.build_tx_data(multisend_txs),
                SafeOperationEnum.DELEGATE_CALL.value,
                0,
                0,
                0,
                None,
                None,
                safe_nonce=safe_nonce,
            )
        else:
            # Multisend not defined
            return None

        safe_tx = self.sign_transaction(safe_tx)
        if not safe_tx.signatures:
            print_formatted_text(
                HTML("<ansired>At least one owner must be loaded</ansired>")
            )
            return None
        else:
            return safe_tx

    def get_signers(self) -> Tuple[List[LocalAccount], List[HwWallet]]:
        """

        :return: Tuple with eoa signers and hw_wallet signers
        """
        permitted_signers = self.get_permitted_signers()
        threshold = self.safe_cli_info.threshold
        eoa_signers: List[Account] = (
            []
        )  # Some accounts that are not an owner can be loaded
        for account in self.accounts:
            if account.address in permitted_signers:
                eoa_signers.append(account)
                threshold -= 1
                if threshold == 0:
                    break
        # If still pending required signatures continue with ledger owners
        hw_wallet_signers = []
        if threshold > 0 and self.hw_wallet_manager.wallets:
            for hw_wallet in self.hw_wallet_manager.wallets:
                if hw_wallet.address in permitted_signers:
                    hw_wallet_signers.append(hw_wallet)
                    threshold -= 1
                    if threshold == 0:
                        break

        if self.require_all_signatures and threshold > 0:
            raise NotEnoughSignatures(threshold)

        return (eoa_signers, hw_wallet_signers)

    # TODO Set sender so we can save gas in that signature
    def sign_transaction(self, safe_tx: SafeTx) -> SafeTx:
        eoa_signers, hw_wallets_signers = self.get_signers()
        for selected_account in eoa_signers:
            safe_tx.sign(selected_account.key)

        # Sign with ledger
        if len(hw_wallets_signers):
            safe_tx = self.hw_wallet_manager.sign_safe_tx(safe_tx, hw_wallets_signers)

        return safe_tx

    @require_tx_service
    def _require_tx_service_mode(self):
        print_formatted_text(
            HTML(
                "<ansired>First enter tx-service mode using <b>tx-service</b> command</ansired>"
            )
        )

    def get_delegates(self):
        return self._require_tx_service_mode()

    def add_delegate(self, delegate_address: str, label: str, signer_address: str):
        return self._require_tx_service_mode()

    def remove_delegate(self, delegate_address: str, signer_address: str):
        return self._require_tx_service_mode()

    def submit_signatures(self, safe_tx_hash: bytes) -> bool:
        return self._require_tx_service_mode()

    def get_balances(self):
        return self._require_tx_service_mode()

    def get_transaction_history(self):
        return self._require_tx_service_mode()

    def batch_txs(self, safe_nonce: int, safe_tx_hashes: Sequence[bytes]) -> bool:
        return self._require_tx_service_mode()

    def execute_tx(self, safe_tx_hash: Sequence[bytes]) -> bool:
        return self._require_tx_service_mode()

    def get_permitted_signers(self) -> Set[ChecksumAddress]:
        """
        :return: Accounts that can sign a transaction
        """
        return set(self.safe_cli_info.owners)

    def drain(self, to: str):
        # Getting all events related with ERC20 transfers
        last = self.ethereum_client.get_block("latest")["number"]
        token_addresses = get_erc_20_list(self.ethereum_client, self.address, 1, last)
        safe_txs = []
        for token_address in token_addresses:
            balance = self.ethereum_client.erc20.get_balance(
                self.address, token_address
            )
            if balance > 0:
                transaction = (
                    get_erc20_contract(self.ethereum_client.w3, token_address)
                    .functions.transfer(to, balance)
                    .build_transaction({"from": self.address, "gas": 0, "gasPrice": 0})
                )

                safe_tx = self.prepare_safe_transaction(
                    token_address,
                    0,
                    HexBytes(transaction["data"]),
                    SafeOperationEnum.CALL,
                    safe_nonce=None,
                )
                safe_txs.append(safe_tx)

        # Getting ethereum balance
        balance_eth = self.ethereum_client.get_balance(self.address)
        if balance_eth:
            safe_tx = self.prepare_safe_transaction(
                to,
                balance_eth,
                b"",
                SafeOperationEnum.CALL,
                safe_nonce=None,
            )
            safe_txs.append(safe_tx)

        if safe_txs:
            multisend_tx = self.batch_safe_txs(self.get_nonce(), safe_txs)
            if multisend_tx is not None:
                if self.execute_safe_transaction(multisend_tx):
                    print_formatted_text(
                        HTML(
                            "<ansigreen>Transaction to drain account correctly executed</ansigreen>"
                        )
                    )
        else:
            print_formatted_text(
                HTML("<ansigreen>Safe account is currently empty</ansigreen>")
            )

    def remove_proposed_transaction(self, safe_tx_hash: bytes):
        return self._require_tx_service_mode()

    def process_command(self, first_command: str, rest_command: List[str]) -> bool:
        if first_command == "help":
            print_formatted_text("I still cannot help you")
        elif first_command == "refresh":
            print_formatted_text("Reloading Safe information")
            self.refresh_safe_cli_info()

        return False


import os
import unittest
from unittest import mock
from unittest.mock import MagicMock

from eth_account import Account
from hexbytes import HexBytes
from trezorlib.client import TrezorClient
from trezorlib.exceptions import Cancelled, OutdatedFirmwareError, PinException
from trezorlib.messages import EthereumTypedDataSignature
from trezorlib.transport import TransportException
from trezorlib.ui import ClickUI

from gnosis.eth.eip712 import eip712_encode
from gnosis.safe import SafeTx
from gnosis.safe.tests.safe_test_case import SafeTestCaseMixin

from safe_cli.operators.exceptions import HardwareWalletException
from safe_cli.operators.hw_wallets.trezor_wallet import TrezorWallet


class TestTrezorManager(SafeTestCaseMixin, unittest.TestCase):
    @mock.patch(
        "safe_cli.operators.hw_wallets.trezor_wallet.get_trezor_client",
        return_value=None,
    )
    @mock.patch(
        "safe_cli.operators.hw_wallets.trezor_wallet.get_address",
        return_value=None,
    )
    def test_setup_trezor_wallet(
        self, mock_trezor_client: MagicMock, mock_get_address: MagicMock
    ):
        trezor_wallet = TrezorWallet("44'/60'/0'/0")
        self.assertIsNone(trezor_wallet.client)

    @mock.patch(
        "safe_cli.operators.hw_wallets.trezor_wallet.sign_typed_data_hash",
        autospec=True,
    )
    @mock.patch(
        "safe_cli.operators.hw_wallets.trezor_wallet.get_address",
        autospec=True,
    )
    @mock.patch(
        "safe_cli.operators.hw_wallets.trezor_wallet.get_trezor_client",
        autospec=True,
    )
    def test_hw_device_exception(
        self,
        mock_trezor_client: MagicMock,
        mock_trezor_get_address: MagicMock,
        mock_trezor_sign: MagicMock,
    ):
        derivation_path = "44'/60'/0'/0"
        transport_mock = MagicMock(auto_spec=True)
        mock_trezor_client.return_value = TrezorClient(
            transport_mock, ui=ClickUI(), _init_device=False
        )
        mock_trezor_client.return_value.is_outdated = MagicMock(return_value=False)
        random_domain_bytes = os.urandom(32)
        random_message_bytes = os.urandom(32)

        mock_trezor_get_address.side_effect = TransportException
        with self.assertRaises(HardwareWalletException):
            TrezorWallet(derivation_path)

        mock_trezor_get_address.side_effect = PinException
        with self.assertRaises(HardwareWalletException):
            TrezorWallet(derivation_path)

        mock_trezor_get_address.side_effect = Cancelled
        with self.assertRaises(HardwareWalletException):
            TrezorWallet(derivation_path)

        mock_trezor_get_address.side_effect = OutdatedFirmwareError
        with self.assertRaises(HardwareWalletException):
            TrezorWallet(derivation_path)

        mock_trezor_get_address.side_effect = None
        mock_trezor_get_address.return_value = Account.create().address
        mock_trezor_sign.side_effect = TransportException
        with self.assertRaises(HardwareWalletException):
            trezor_wallet = TrezorWallet(derivation_path)
            trezor_wallet.sign_typed_hash(random_domain_bytes, random_message_bytes)

        mock_trezor_sign.side_effect = PinException
        with self.assertRaises(HardwareWalletException):
            trezor_wallet = TrezorWallet(derivation_path)
            trezor_wallet.sign_typed_hash(random_domain_bytes, random_message_bytes)

        mock_trezor_sign.side_effect = Cancelled
        with self.assertRaises(HardwareWalletException):
            trezor_wallet = TrezorWallet(derivation_path)
            trezor_wallet.sign_typed_hash(random_domain_bytes, random_message_bytes)

        mock_trezor_sign.side_effect = OutdatedFirmwareError
        with self.assertRaises(HardwareWalletException):
            trezor_wallet = TrezorWallet(derivation_path)
            trezor_wallet.sign_typed_hash(random_domain_bytes, random_message_bytes)

    @mock.patch(
        "safe_cli.operators.hw_wallets.trezor_wallet.get_address",
        autospec=True,
    )
    @mock.patch(
        "safe_cli.operators.hw_wallets.trezor_wallet.get_trezor_client",
        autospec=True,
    )
    def test_sign_typed_hash(
        self, mock_trezor_client: MagicMock, mock_get_address: MagicMock
    ):
        owner = Account.create()
        to = Account.create()
        transport_mock = MagicMock(auto_spec=True)
        mock_trezor_client.return_value = TrezorClient(
            transport_mock, ui=ClickUI(), _init_device=False
        )
        mock_trezor_client.return_value.is_outdated = MagicMock(return_value=False)
        mock_get_address.return_value = owner.address
        trezor_wallet = TrezorWallet("44'/60'/0'/0")

        safe = self.deploy_test_safe(
            owners=[owner.address],
            threshold=1,
            initial_funding_wei=self.w3.to_wei(0.1, "ether"),
        )
        safe_tx = SafeTx(
            self.ethereum_client,
            safe.address,
            to.address,
            10,
            b"",
            0,
            200000,
            200000,
            self.gas_price,
            None,
            None,
            safe_nonce=0,
        )
        encode_hash = eip712_encode(safe_tx.eip712_structured_data)
        expected_signature = safe_tx.sign(owner.key)

        trezor_return_signature = EthereumTypedDataSignature(
            signature=expected_signature, address=trezor_wallet.address
        )
        mock_trezor_client.return_value.call = MagicMock(
            return_value=trezor_return_signature
        )
        signature = trezor_wallet.sign_typed_hash(encode_hash[1], encode_hash[2])
        self.assertEqual(expected_signature, signature)

    @mock.patch(
        "safe_cli.operators.hw_wallets.trezor_wallet.sign_tx",
        autospec=True,
    )
    @mock.patch(
        "safe_cli.operators.hw_wallets.trezor_wallet.sign_tx_eip1559",
        autospec=True,
    )
    @mock.patch(
        "safe_cli.operators.hw_wallets.trezor_wallet.get_address",
        autospec=True,
    )
    @mock.patch(
        "safe_cli.operators.hw_wallets.trezor_wallet.get_trezor_client",
        autospec=True,
    )
    def test_get_signed_raw_transaction(
        self,
        mock_trezor_client: MagicMock,
        mock_get_address: MagicMock,
        mock_sign_tx_eip1559: MagicMock,
        mock_sign_tx: MagicMock,
    ):
        owner = Account.create()
        to = Account.create()
        transport_mock = MagicMock(auto_spec=True)
        mock_trezor_client.return_value = TrezorClient(
            transport_mock, ui=ClickUI(), _init_device=False
        )
        mock_trezor_client.return_value.is_outdated = MagicMock(return_value=False)
        mock_get_address.return_value = owner.address
        trezor_wallet = TrezorWallet("44'/60'/0'/0")

        safe = self.deploy_test_safe(
            owners=[owner.address],
            threshold=1,
            initial_funding_wei=self.w3.to_wei(0.1, "ether"),
        )
        safe_tx = SafeTx(
            self.ethereum_client,
            safe.address,
            to.address,
            10,
            b"",
            0,
            200000,
            200000,
            self.gas_price,
            None,
            None,
            safe_nonce=0,
        )
        safe_tx.sign(owner.key)
        # Legacy transaction
        tx_parameters = {
            "from": owner.address,
            "gasPrice": safe_tx.w3.eth.gas_price,
            "nonce": 0,
            "gas": safe_tx.recommended_gas(),
        }
        safe_tx.tx = safe_tx.w3_tx.build_transaction(tx_parameters)
        signed_fields = safe_tx.w3.eth.account.sign_transaction(
            safe_tx.tx, private_key=owner.key
        )

        mock_sign_tx.return_value = (
            HexBytes(signed_fields.v),
            HexBytes(signed_fields.r),
            HexBytes(signed_fields.s),
        )

        raw_signed_tx = trezor_wallet.get_signed_raw_transaction(
            safe_tx.tx, safe_tx.ethereum_client.get_chain_id()
        )  # return raw signed transaction
        mock_sign_tx.assert_called_once()
        self.assertEqual(signed_fields.rawTransaction, HexBytes(raw_signed_tx))

        # EIP1559 transaction
        tx_parameters = {
            "from": owner.address,
            "maxPriorityFeePerGas": safe_tx.w3.eth.gas_price,
            "maxFeePerGas": safe_tx.w3.eth.gas_price,
            "nonce": 1,
            "gas": safe_tx.recommended_gas(),
        }
        safe_tx.tx = safe_tx.w3_tx.build_transaction(tx_parameters)
        signed_fields = safe_tx.w3.eth.account.sign_transaction(
            safe_tx.tx, private_key=owner.key
        )

        mock_sign_tx_eip1559.return_value = (
            signed_fields.v,
            HexBytes(signed_fields.r),
            HexBytes(signed_fields.s),
        )
        raw_signed_tx = trezor_wallet.get_signed_raw_transaction(
            safe_tx.tx, safe_tx.ethereum_client.get_chain_id()
        )  # return raw signed transaction
        mock_sign_tx_eip1559.assert_called_once()
        self.assertEqual(signed_fields.rawTransaction, HexBytes(raw_signed_tx))



from eth_account.signers.local import LocalAccount

from gnosis.eth.tests.utils import deploy_erc20, send_tx

from safe_cli.operators import SafeOperator


# Deploy ERC20 tokens and send it to Safe account
def generate_transfers_erc20(
    w3,
    safe_operator: SafeOperator,
    account: LocalAccount,
    num_contracts: int = 3,
    num_tokens_to_send: int = 3,
):

    for i in range(num_contracts):
        token_name = "MOI{i}"
        # Deploy ERC20
        erc20_contract = deploy_erc20(
            safe_operator.safe.w3,
            account,
            token_name,
            token_name,
            account.address,
            10,
        )
        # Create a transaction of 1 token ERC20
        transaction = erc20_contract.functions.transfer(
            safe_operator.address, 1
        ).build_transaction({"from": account.address})
        nonce = w3.eth.get_transaction_count(
            account.address, block_identifier="pending"
        )
        # Execute one transfer per each token to send
        for i in range(num_tokens_to_send):
            transaction["nonce"] = nonce
            send_tx(w3, transaction, account)
            nonce += 1



import argparse
import unittest

from eth_account import Account
from hexbytes import HexBytes

from safe_cli.argparse_validators import (
    check_ethereum_address,
    check_hex_str,
    check_keccak256_hash,
    check_private_key,
)
from safe_cli.safe_creator import check_positive_integer

from .safe_cli_test_case_mixin import SafeCliTestCaseMixin


class TestArgparseValidators(SafeCliTestCaseMixin, unittest.TestCase):
    def test_check_positive_integer(self):
        self.assertEqual(check_positive_integer(1), 1)
        self.assertEqual(check_positive_integer(500), 500)

        with self.assertRaises(argparse.ArgumentTypeError):
            self.assertEqual(check_positive_integer(0), 0)
        with self.assertRaises(argparse.ArgumentTypeError):
            check_positive_integer(-1)

    def test_check_ethereum_address(self):
        address = "0x4127839cdf4F73d9fC9a2C2861d8d1799e9DF40C"
        self.assertEqual(check_ethereum_address(address), address)

        not_valid_address = "0x4127839CDf4F73d9fC9a2C2861d8d1799e9DF40C"
        with self.assertRaises(argparse.ArgumentTypeError):
            check_ethereum_address(not_valid_address)

    def test_check_private_key(self):
        account = Account.create()
        self.assertEqual(check_private_key(account.key.hex()), account.key.hex())

        with self.assertRaises(argparse.ArgumentTypeError):
            check_private_key("Random")

    def test_check_hex_str(self):
        self.assertEqual(check_hex_str("0x12"), HexBytes("0x12"))

        with self.assertRaises(argparse.ArgumentTypeError):
            check_hex_str("0x12x")

    def test_check_keccak256_hash(self):
        random_hash = (
            "0x8aca9664752dbae36135fd0956c956fc4a370feeac67485b49bcd4b99608ae41"
        )
        self.assertEqual(check_keccak256_hash(random_hash), HexBytes(random_hash))

        with self.assertRaisesRegex(
            argparse.ArgumentTypeError,
            f"{random_hash[:-2]} is not a valid keccak256 hash hexadecimal string",
        ):
            # Size must match
            check_keccak256_hash(random_hash[:-2])

        with self.assertRaisesRegex(
            argparse.ArgumentTypeError, "0x12x is not a valid hexadecimal string"
        ):
            check_keccak256_hash("0x12x")


if __name__ == "__main__":
    unittest.main()



import unittest

from prompt_toolkit.completion import CompleteEvent, Completion
from prompt_toolkit.document import Document

from safe_cli.safe_completer import SafeCompleter


class TestSafeCompleter(unittest.TestCase):
    def test_get_completions(self):
        safe_completer = SafeCompleter()

        with self.assertRaises(StopIteration):
            next(
                safe_completer.get_completions(
                    Document("not-supported-command"), CompleteEvent()
                )
            )

        self.assertIsInstance(
            next(
                safe_completer.get_completions(Document("send_ether"), CompleteEvent())
            ),
            Completion,
        )


if __name__ == "__main__":
    unittest.main()



import os
from typing import List, Optional

from eth_typing import ChecksumAddress
from prompt_toolkit import HTML, print_formatted_text

from gnosis.eth import EthereumClient
from gnosis.safe.api import TransactionServiceApi


def get_erc_20_list(
    ethereum_client: EthereumClient,
    safe_address: str,
    from_block: int,
    to_block: int,
    block_step: int = 500000,
) -> list:
    """

    :param ethereum_client:
    :param safe_address:
    :param from_block:
    :param to_block:
    :param block_step: is the number of blocks retrieved for each get until get all blocks between from_block until to_block
    :return: a list of address of ERC20 tokens related with the safe_address
    """
    addresses = set()
    for i in range(from_block, to_block + 1, block_step):
        events = ethereum_client.erc20.get_total_transfer_history(
            from_block=i, to_block=i + (block_step - 1), addresses=[safe_address]
        )
        for event in events:
            if "value" in event["args"]:
                addresses.add(event["address"])

    return addresses


def get_input(*args, **kwargs):
    return input(*args, **kwargs)


def yes_or_no_question(question: str, default_no: bool = True) -> bool:
    if "PYTEST_CURRENT_TEST" in os.environ:
        return True  # Ignore confirmations when running tests

    choices = " [y/N]: " if default_no else " [Y/n]: "
    default_answer = "n" if default_no else "y"
    reply = str(get_input(question + choices)).lower().strip() or default_answer
    if reply[0] == "y":
        return True
    if reply[0] == "n":
        return False
    else:
        return False if default_no else True


def choose_option_from_list(
    question: str, options: List, default_option: int = 0
) -> Optional[int]:
    if "PYTEST_CURRENT_TEST" in os.environ:
        return default_option  # Ignore confirmations when running tests
    number_options = len(options)
    for number_option, option in enumerate(options):
        print_formatted_text(HTML(f"{number_option} - <b>{option}</b> "))
    choices = f" [0-{number_options-1}] default {default_option}: "
    reply = str(get_input(question + choices)).lower().strip() or str(default_option)
    try:
        option = int(reply)
    except ValueError:
        print_formatted_text(HTML("<ansired> Option must be an integer </ansired>"))
        return None

    if option not in range(0, number_options):
        print_formatted_text(
            HTML(f"<ansired> {option} is not between [0-{number_options-1}] </ansired>")
        )
        return None

    return option


def get_safe_from_owner(
    owner: ChecksumAddress, node_url: str
) -> Optional[ChecksumAddress]:
    """
    Show a list of Safe to chose between them and return the selected one.
    :param owner:
    :param node_url:
    :return: Safe address of a selected Safe
    """
    ethereum_client = EthereumClient(node_url)
    safe_tx_service = TransactionServiceApi.from_ethereum_client(ethereum_client)
    safes = safe_tx_service.get_safes_for_owner(owner)
    if safes:
        option = choose_option_from_list(
            "Select the Safe to initialize the safe-cli", safes
        )
        if option is not None:
            return safes[option]
    else:
        raise ValueError(f"No safe was found for the specified owner {owner}")



import unittest

from eth_account import Account
from web3 import Web3

from gnosis.safe import Safe

from safe_cli.operators.safe_operator import SafeOperator
from safe_cli.prompt_parser import PromptParser

from .safe_cli_test_case_mixin import SafeCliTestCaseMixin


class SafeCliTestCase(SafeCliTestCaseMixin, unittest.TestCase):
    def test_safe_cli_happy_path(self):
        accounts = [self.ethereum_test_account, Account.create()]
        account_addresses = [account.address for account in accounts]
        safe_address = self.deploy_test_safe(
            owners=account_addresses,
            threshold=2,
            initial_funding_wei=self.w3.to_wei(1, "ether"),
        ).address
        safe = Safe(safe_address, self.ethereum_client)
        safe_operator = SafeOperator(safe_address, self.ethereum_node_url)
        prompt_parser = PromptParser(safe_operator)
        random_address = Account.create().address

        self.assertEqual(safe_operator.accounts, set())
        prompt_parser.process_command(
            f"load_cli_owners {self.ethereum_test_account.key.hex()}"
        )
        self.assertEqual(safe_operator.default_sender, self.ethereum_test_account)
        self.assertEqual(safe_operator.accounts, {self.ethereum_test_account})

        prompt_parser.process_command(
            f"send_ether {random_address} 1"
        )  # No enough signatures
        self.assertEqual(self.ethereum_client.get_balance(random_address), 0)

        value = 123
        prompt_parser.process_command(f"load_cli_owners {accounts[1].key.hex()}")
        prompt_parser.process_command(f"send_ether {random_address} {value}")
        self.assertEqual(self.ethereum_client.get_balance(random_address), value)

        # Change threshold
        self.assertEqual(safe_operator.safe_cli_info.threshold, 2)
        self.assertEqual(safe.retrieve_threshold(), 2)
        prompt_parser.process_command("change_threshold 1")
        self.assertEqual(safe_operator.safe_cli_info.threshold, 1)
        self.assertEqual(safe.retrieve_threshold(), 1)

        # Approve Hash
        safe_tx_hash = Web3.keccak(text="hola")
        self.assertFalse(
            safe_operator.safe.retrieve_is_hash_approved(
                accounts[0].address, safe_tx_hash
            )
        )
        prompt_parser.process_command(
            f"approve_hash {safe_tx_hash.hex()} {accounts[0].address}"
        )
        self.assertTrue(
            safe_operator.safe.retrieve_is_hash_approved(
                accounts[0].address, safe_tx_hash
            )
        )

        # Remove owner
        self.assertEqual(len(safe_operator.safe_cli_info.owners), 2)
        self.assertEqual(len(safe.retrieve_owners()), 2)
        prompt_parser.process_command(f"remove_owner {accounts[1].address}")
        self.assertEqual(
            safe_operator.safe_cli_info.owners, [self.ethereum_test_account.address]
        )
        self.assertEqual(safe.retrieve_owners(), [self.ethereum_test_account.address])


if __name__ == "__main__":
    unittest.main()



import unittest

from eth_utils import ValidationError

from safe_cli.ethereum_hd_wallet import get_account_from_words, get_address_from_words


class TestEthereumHdWallet(unittest.TestCase):
    def test_get_account_from_words(self):
        words = (
            "loan satoshi action taste party limit cat elder powder dress link decline"
        )

        expected = [
            (
                "0x9c8a7003407957Adee0e70f3094aDA208FDd4CF1",
                b"\x9e!\xaf\xeaoe\xc5l\xf0\x8c\xde~s\xaa\xb3O\x14\xb0&\x97|N\xf5l\x8a#\x14\xe9\xcc\xd2\x86%",
            ),
            (
                "0x09A512D5ecfF9492Cb0f50AFF853dFD8B4ec3EB1",
                b"\xdd\xb7\xf84\xa2\xea\x03Cd\xce\xcf3i\xcb\xf1\xa7\xe0\xa5\x0b\xc3-\xc0\xc1 D\xe0&!\xbb\xa9\x85\xf9",
            ),
            (
                "0x58A0494D85E36f7DFc7d3BA56681ab6efDEf0A19",
                b"\xf8\xfd*\x95\t\x0e\xff\x9c?\rc\x8f*\xb2x\xe7<\xb5\xaa\xf87\xba\xe8PT\x0b\xcf(\xa42\xc7O",
            ),
            (
                "0xa955FBEFe163C6BeAb64A7E854612639782b797F",
                b"h<\xa0\xdf\xec\xf7\xe9\xfe!!z\xfav\xd1(\xa4;\x03V\x84\xb5\x0fE\x06\x90\x8a\x1a\x01\xd8\x7f\xf8o",
            ),
            (
                "0x38DF5615C090Ca80930986295B9bbFCc36a0Ae7B",
                b"h\xdb-\xc1\xbf\xc3\xf0\x02\xb3\x18\xd7\x81BE~"
                b"\xb6\x1d\xa9\x88@\xfd\x80\x82~\xcf\x1c\x08\x17\x96\x1f\xf0\xed",
            ),
        ]

        for index in range(len(expected)):
            account = get_account_from_words(words, index=index)
            self.assertEqual(account.address, expected[index][0])
            self.assertEqual(account.key, expected[index][1])

    def test_get_account_from_words_exception(self):
        words = "loan satoshi action taste party limit cat elder powder decline"  # Not valid
        with self.assertRaises(ValidationError):
            get_account_from_words(words)

        with self.assertRaises(ValidationError):
            get_account_from_words("")

    def test_get_address_from_words(self):
        words = (
            "loan satoshi action taste party limit cat elder powder dress link decline"
        )

        expected = [
            "0x9c8a7003407957Adee0e70f3094aDA208FDd4CF1",
            "0x09A512D5ecfF9492Cb0f50AFF853dFD8B4ec3EB1",
            "0x58A0494D85E36f7DFc7d3BA56681ab6efDEf0A19",
            "0xa955FBEFe163C6BeAb64A7E854612639782b797F",
            "0x38DF5615C090Ca80930986295B9bbFCc36a0Ae7B",
        ]

        for index in range(len(expected)):
            self.assertEqual(
                get_address_from_words(words, index=index), expected[index]
            )


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from unittest import mock
from unittest.mock import MagicMock

from eth_account import Account

from safe_cli.utils import choose_option_from_list, yes_or_no_question


class TestUtils(unittest.TestCase):
    @mock.patch("safe_cli.utils.get_input")
    def test_yes_or_no_question(self, input_mock: MagicMock):
        # Input defaults to `yes` if running tests
        pytest_current_test = os.environ.pop("PYTEST_CURRENT_TEST")
        input_mock.return_value = "yes"
        self.assertTrue(yes_or_no_question(""))

        input_mock.return_value = "yay"
        self.assertTrue(yes_or_no_question(""))

        input_mock.return_value = "Y"
        self.assertTrue(yes_or_no_question(""))

        input_mock.return_value = "Nope"
        self.assertFalse(yes_or_no_question(""))

        input_mock.return_value = "No"
        self.assertFalse(yes_or_no_question(""))

        input_mock.return_value = "n"
        self.assertFalse(yes_or_no_question(""))

        input_mock.return_value = "random"
        self.assertFalse(yes_or_no_question(""))

        os.environ["PYTEST_CURRENT_TEST"] = pytest_current_test

    @mock.patch("safe_cli.utils.get_input")
    def test_choose_option_from_list(self, input_mock: MagicMock):
        pytest_current_test = os.environ.pop("PYTEST_CURRENT_TEST")
        address = Account.create().address
        options = [address for i in range(0, 5)]
        input_mock.return_value = ""
        self.assertEqual(choose_option_from_list("", options), 0)
        input_mock.return_value = ""
        self.assertEqual(choose_option_from_list("", options, 4), 4)
        input_mock.return_value = "m"
        self.assertIsNone(choose_option_from_list("", options))
        input_mock.return_value = "10"
        self.assertIsNone(choose_option_from_list("", options))
        input_mock.return_value = "1"
        self.assertEqual(choose_option_from_list("", options), 1)

        os.environ["PYTEST_CURRENT_TEST"] = pytest_current_test


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest import mock
from unittest.mock import MagicMock

from eth_account import Account
from hexbytes import HexBytes
from ledgereth.objects import LedgerAccount
from web3 import Web3

from gnosis.eth import EthereumClient
from gnosis.safe import SafeTx
from gnosis.safe.api import TransactionServiceApi

from safe_cli.operators import SafeOperatorMode, SafeTxServiceOperator

from .mocks.balances_mock import balances_mock
from .mocks.data_decoded_mock import data_decoded_mock
from .mocks.multisig_tx_mock import GetMultisigTxRequestMock
from .mocks.txs_mock import txs_mock
from .safe_cli_test_case_mixin import SafeCliTestCaseMixin


class TestSafeTxServiceOperator(SafeCliTestCaseMixin, unittest.TestCase):
    def test_setup_operator(self):
        safe_operator = self.setup_operator(
            number_owners=1, mode=SafeOperatorMode.TX_SERVICE
        )
        self.assertIsInstance(safe_operator, SafeTxServiceOperator)

    def test_approve_hash(self):
        safe_operator = self.setup_operator(
            number_owners=1, mode=SafeOperatorMode.TX_SERVICE
        )
        safe_tx_hash = Web3.keccak(text="random-test")
        with self.assertRaises(NotImplementedError):
            safe_operator.approve_hash(
                safe_tx_hash, list(safe_operator.accounts)[0].address
            )

    @mock.patch.object(TransactionServiceApi, "get_delegates")
    def test_get_delegates(self, get_delegates_mock: MagicMock):
        get_delegates_mock.return_value = [
            {
                "safe": "0x172766BDBf2e05405a890BBfa120C726CA862a6c",
                "delegate": "0xFACA096a40a68557516202B75447c7494c83b522",
                "delegator": "0xaf2F9D5D7991Ba570203f2bf2Ff7ecC90d4B906D",
                "label": "Testing",
            },
            {
                "safe": "0x172766BDBf2e05405a890BBfa120C726CA862a6c",
                "delegate": "0xFACA096a40a68557516202B75447c7494c83b522",
                "delegator": "0xd6e1eF1B88BE12b39785d24aD21235FcF48C6409",
                "label": "Testing",
            },
        ]
        safe_operator = self.setup_operator(
            number_owners=1, mode=SafeOperatorMode.TX_SERVICE
        )
        self.assertEqual(
            safe_operator.get_delegates(),
            [
                [
                    "0xFACA096a40a68557516202B75447c7494c83b522",
                    "0xaf2F9D5D7991Ba570203f2bf2Ff7ecC90d4B906D",
                    "Testing",
                ],
                [
                    "0xFACA096a40a68557516202B75447c7494c83b522",
                    "0xd6e1eF1B88BE12b39785d24aD21235FcF48C6409",
                    "Testing",
                ],
            ],
        )

    @mock.patch.object(TransactionServiceApi, "add_delegate", return_value=None)
    def test_add_delegate(self, add_delegate_mock: MagicMock):
        safe_operator = self.setup_operator(
            number_owners=1, mode=SafeOperatorMode.TX_SERVICE
        )
        delegate_address = Account.create().address
        label = "Test"
        signer = list(safe_operator.accounts)[0]
        self.assertTrue(
            safe_operator.add_delegate(delegate_address, label, signer.address)
        )
        add_delegate_mock.assert_called_with(
            safe_operator.address, delegate_address, label, signer
        )

    @mock.patch.object(TransactionServiceApi, "remove_delegate", return_value=None)
    def test_remove_delegate(self, remove_delegate_mock: MagicMock):
        safe_operator = self.setup_operator(
            number_owners=1, mode=SafeOperatorMode.TX_SERVICE
        )
        delegate_address = Account.create().address
        signer = list(safe_operator.accounts)[0]
        self.assertTrue(safe_operator.remove_delegate(delegate_address, signer.address))
        remove_delegate_mock.assert_called_with(
            safe_operator.address, delegate_address, signer
        )

    @mock.patch.object(TransactionServiceApi, "delete_transaction", return_value=None)
    @mock.patch.object(TransactionServiceApi, "get_safe_transaction")
    @mock.patch(
        "gnosis.safe.api.transaction_service_api.transaction_service_messages.get_totp",
        return_value=8365,
    )
    def test_remove_transaction(
        self,
        mock_get_totp,
        get_transaction_mock: MagicMock,
        remove_transaction_mock: MagicMock,
    ):
        safe_operator = self.setup_operator(
            number_owners=1, mode=SafeOperatorMode.TX_SERVICE
        )
        safe_operator.address = "0x23793e7Ce4f2Fd31C993893f658181fD39fB5dEb"
        signer = list(safe_operator.accounts)[0]
        safe_tx_mock = MagicMock()
        get_transaction_mock.return_value = (safe_tx_mock, None)
        safe_tx_hash = HexBytes(
            "0x07eeea19b7d005b561f367e714bf65734d0ecc477de3e8b1fbc20ccb18c8747b"
        )
        expected_signature = "0xbc305f59a62c5bbb002645f658ee62cfa1966f20aca4dc1922b813e9d41bf1f02abd356891a3725af2066dd1758c930574298b4ba180117dc6146bbc4369cc0c1c"
        safe_tx_mock.proposer = signer.address
        self.assertTrue(safe_operator.remove_proposed_transaction(safe_tx_hash))
        remove_transaction_mock.assert_called_with(
            safe_tx_hash.hex(), expected_signature
        )
        # Different proposer should return False
        safe_tx_mock.proposer = Account.create().address
        self.assertFalse(safe_operator.remove_proposed_transaction(safe_tx_hash))

    @mock.patch.object(SafeTxServiceOperator, "get_permitted_signers", return_value=[])
    @mock.patch.object(
        SafeTx, "safe_version", return_value="1.4.1", new_callable=mock.PropertyMock
    )
    @mock.patch.object(TransactionServiceApi, "post_signatures", return_value=None)
    @mock.patch.object(TransactionServiceApi, "_get_request")
    def test_submit_signatures(
        self,
        get_safe_transaction_mock: MagicMock,
        post_signatures_mock: MagicMock,
        safe_version_mock: mock.PropertyMock,
        get_permitted_signers_mock: MagicMock,
    ):
        get_safe_transaction_mock.return_value = GetMultisigTxRequestMock(
            executed=False
        )
        post_signatures_mock.return_value = None

        safe_tx_hash = HexBytes(
            "0xeb5fa8e85dd530397172da07792c5d05dff9ffe5816fc0a260d672e924825b01"
        )
        safe_operator = self.setup_operator(
            number_owners=1, mode=SafeOperatorMode.TX_SERVICE
        )
        # No suitable signers
        self.assertFalse(safe_operator.submit_signatures(safe_tx_hash))
        get_permitted_signers_mock.return_value = {
            list(safe_operator.accounts)[0].address
        }
        # Now signatures must be submited
        self.assertTrue(safe_operator.submit_signatures(safe_tx_hash))

        # Cannot sign executed transactions
        get_safe_transaction_mock.return_value = GetMultisigTxRequestMock(executed=True)
        self.assertFalse(safe_operator.submit_signatures(safe_tx_hash))

        # Test ledger signers
        with mock.patch.object(
            SafeTx, "signers", return_value=["signer"], new_callable=mock.PropertyMock
        ) as mock_safe_tx:
            safe_operator.hw_wallet_manager.sign_safe_tx = MagicMock(
                return_value=mock_safe_tx
            )
            get_safe_transaction_mock.return_value = GetMultisigTxRequestMock(
                executed=False
            )
            safe_operator.hw_wallet_manager.wallets.add(
                LedgerAccount("44'/60'/0'/0", Account.create().address)
            )
            get_permitted_signers_mock.return_value = {
                list(safe_operator.hw_wallet_manager.wallets)[0].address
            }
            self.assertTrue(safe_operator.submit_signatures(safe_tx_hash))

    @mock.patch.object(TransactionServiceApi, "post_transaction", return_value=True)
    @mock.patch.object(
        SafeTx, "safe_version", return_value="1.4.1", new_callable=mock.PropertyMock
    )
    @mock.patch.object(EthereumClient, "is_contract", return_value=True)
    @mock.patch.object(TransactionServiceApi, "_get_request")
    def test_batch_txs(
        self,
        get_safe_transaction_mock: MagicMock,
        is_contract_mock: MagicMock,
        safe_version_mock: mock.PropertyMock,
        post_transaction_mock: MagicMock,
    ):
        safe_operator = self.setup_operator(
            number_owners=1, mode=SafeOperatorMode.TX_SERVICE
        )
        get_safe_transaction_mock.return_value = GetMultisigTxRequestMock(
            executed=False
        )
        safe_tx_hash = HexBytes(
            "0xeb5fa8e85dd530397172da07792c5d05dff9ffe5816fc0a260d672e924825b01"
        )
        safe_nonce = 0
        safe_operator.batch_txs(safe_nonce, [safe_tx_hash])

    @mock.patch.object(
        TransactionServiceApi, "get_balances", return_value=balances_mock
    )
    def test_get_balances(
        self,
        get_balances_mock: MagicMock,
    ):
        safe_operator = self.setup_operator(
            number_owners=1, mode=SafeOperatorMode.TX_SERVICE
        )
        expected = [
            ["ETHER", "0.17100", "Ξ", 18, ""],
            [
                "Basic Attention Token",
                "1.00000",
                "BAT",
                18,
                "0x70cBa46d2e933030E2f274AE58c951C800548AeF",
            ],
            [
                " best tron wallet | tronscan | tronlink | bttc",
                "8888.00000",
                "tronAddress.org",
                0,
                "0x4644A362B7041aff2CF792fD61671b9F75234C27",
            ],
            [
                "CoW Protocol Token",
                "7.45545",
                "COW",
                18,
                "0x91056D4A53E1faa1A84306D4deAEc71085394bC8",
            ],
            [
                "CoW Protocol Token",
                "8.04516",
                "COW",
                18,
                "0x3430d04E42a722c5Ae52C5Bffbf1F230C2677600",
            ],
            [
                "Dai",
                "1000.00000",
                "DAI",
                18,
                "0xdc31Ee1784292379Fbb2964b3B9C4124D8F89C60",
            ],
            [
                "Elden Ring Token",
                "10000.00000",
                "ERT",
                18,
                "0x64685e9733d57Ccc3870ceD50ed9B1f576Fd61B1",
            ],
            [
                "ethAddress.io",
                "888888.00000",
                "ethAddress.io",
                0,
                "0xD14DD51BE76e6dC28756a442dF786845e03a1D60",
            ],
            [
                "GiantETHLP",
                "0.00100",
                "gETH",
                18,
                "0x4c6FFFda6b732b495c974Fb284e9bB38b9f061C8",
            ],
            [
                "Gnosis Token",
                "0.40000",
                "GNO",
                18,
                "0x02ABBDbAaa7b1BB64B5c878f7ac17f8DDa169532",
            ],
            [
                "Liquid staked Ether 2.0",
                "0.03020",
                "stETH",
                18,
                "0x1643E812aE58766192Cf7D2Cf9567dF2C37e9B7F",
            ],
            [
                "RareTron.io",
                "66666.00000",
                "RareTron.io",
                0,
                "0x54FA517F05e11Ffa87f4b22AE87d91Cec0C2D7E1",
            ],
            [
                "Safe Token",
                "10362.07000",
                "SAFE",
                18,
                "0x61fD3b6d656F39395e32f46E2050953376c3f5Ff",
            ],
            [
                "Super ETH",
                "0.00500",
                "ETHx",
                18,
                "0x5943F705aBb6834Cad767e6E4bB258Bc48D9C947",
            ],
            [
                "Uniswap",
                "0.02130",
                "UNI",
                18,
                "0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984",
            ],
            [
                "USD Coin",
                "10.00000",
                "USDC",
                6,
                "0xD87Ba7A50B2E7E660f678A895E4B72E7CB4CCd9C",
            ],
            [
                "USD Coin",
                "0.92509",
                "USDC",
                6,
                "0xe0C9275E44Ea80eF17579d33c55136b7DA269aEb",
            ],
            [
                "VanityTron.io",
                "6666666.00000",
                "VanityTron.io",
                0,
                "0x38d9639da06f7a726cAf40c2E1010ac88517A085",
            ],
            [
                "VanityTRX.org",
                "888888.00000",
                "VanityTRX.org",
                0,
                "0x1B809925ba90c541d895D19f0b7D70eE281a987F",
            ],
            [
                "Wrapped Ether",
                "0.00050",
                "WETH",
                18,
                "0xdFCeA9088c8A88A76FF74892C1457C17dfeef9C1",
            ],
            [
                "Wrapped Ether",
                "0.00560",
                "WETH",
                18,
                "0xB4FBF271143F4FBf7B91A5ded31805e42b2208d6",
            ],
        ]
        self.assertEqual(safe_operator.get_balances(), expected)

    @mock.patch.object(TransactionServiceApi, "get_transactions", return_value=txs_mock)
    def test_get_transaction_history(
        self,
        get_transactions_mock: MagicMock,
    ):
        safe_operator = self.setup_operator(
            number_owners=1, mode=SafeOperatorMode.TX_SERVICE
        )
        expected = [
            ["ETHER", "0.17100", "Ξ", 18, ""],
            [
                "Basic Attention Token",
                "1.00000",
                "BAT",
                18,
                "0x70cBa46d2e933030E2f274AE58c951C800548AeF",
            ],
            [
                " best tron wallet | tronscan | tronlink | bttc",
                "8888.00000",
                "tronAddress.org",
                0,
                "0x4644A362B7041aff2CF792fD61671b9F75234C27",
            ],
            [
                "CoW Protocol Token",
                "7.45545",
                "COW",
                18,
                "0x91056D4A53E1faa1A84306D4deAEc71085394bC8",
            ],
            [
                "CoW Protocol Token",
                "8.04516",
                "COW",
                18,
                "0x3430d04E42a722c5Ae52C5Bffbf1F230C2677600",
            ],
            [
                "Dai",
                "1000.00000",
                "DAI",
                18,
                "0xdc31Ee1784292379Fbb2964b3B9C4124D8F89C60",
            ],
            [
                "Elden Ring Token",
                "10000.00000",
                "ERT",
                18,
                "0x64685e9733d57Ccc3870ceD50ed9B1f576Fd61B1",
            ],
            [
                "ethAddress.io",
                "888888.00000",
                "ethAddress.io",
                0,
                "0xD14DD51BE76e6dC28756a442dF786845e03a1D60",
            ],
            [
                "GiantETHLP",
                "0.00100",
                "gETH",
                18,
                "0x4c6FFFda6b732b495c974Fb284e9bB38b9f061C8",
            ],
            [
                "Gnosis Token",
                "0.40000",
                "GNO",
                18,
                "0x02ABBDbAaa7b1BB64B5c878f7ac17f8DDa169532",
            ],
            [
                "Liquid staked Ether 2.0",
                "0.03020",
                "stETH",
                18,
                "0x1643E812aE58766192Cf7D2Cf9567dF2C37e9B7F",
            ],
            [
                "RareTron.io",
                "66666.00000",
                "RareTron.io",
                0,
                "0x54FA517F05e11Ffa87f4b22AE87d91Cec0C2D7E1",
            ],
            [
                "Safe Token",
                "10362.07000",
                "SAFE",
                18,
                "0x61fD3b6d656F39395e32f46E2050953376c3f5Ff",
            ],
            [
                "Super ETH",
                "0.00500",
                "ETHx",
                18,
                "0x5943F705aBb6834Cad767e6E4bB258Bc48D9C947",
            ],
            [
                "Uniswap",
                "0.02130",
                "UNI",
                18,
                "0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984",
            ],
            [
                "USD Coin",
                "10.00000",
                "USDC",
                6,
                "0xD87Ba7A50B2E7E660f678A895E4B72E7CB4CCd9C",
            ],
            [
                "USD Coin",
                "0.92509",
                "USDC",
                6,
                "0xe0C9275E44Ea80eF17579d33c55136b7DA269aEb",
            ],
            [
                "VanityTron.io",
                "6666666.00000",
                "VanityTron.io",
                0,
                "0x38d9639da06f7a726cAf40c2E1010ac88517A085",
            ],
            [
                "VanityTRX.org",
                "888888.00000",
                "VanityTRX.org",
                0,
                "0x1B809925ba90c541d895D19f0b7D70eE281a987F",
            ],
            [
                "Wrapped Ether",
                "0.00050",
                "WETH",
                18,
                "0xdFCeA9088c8A88A76FF74892C1457C17dfeef9C1",
            ],
            [
                "Wrapped Ether",
                "0.00560",
                "WETH",
                18,
                "0xB4FBF271143F4FBf7B91A5ded31805e42b2208d6",
            ],
        ]
        expected = [
            [
                "\x1b[0m\x1b[1m\x1b[32m213",
                "0x4127839cdf4F73d9fC9a2C2861d8d1799e9DF40C",
                "100000",
                "0x6d1de5637d9b78a498145e5bcf13219ac88bb2fef8753887722b31c4290c2bf4",
                "0xf10a3c13a7b8cac3e343ca5f2f328c15b1aeb6031a7b1cbc4b9a65547d4d7070",
            ],
            [
                "\x1b[0m\x1b[32m212",
                "0xB4FBF271143F4FBf7B91A5ded31805e42b2208d6",
                "0",
                "0xf59007f3240cada5a9514aee3eb9de4d84bd8de8ff6507f26d6b309a8b07ee4b",
                "0xa26abcadcad3176b51f91bc0bd4c4c35d9afe1db5467268642452e2911316606",
                "transfer: 0xc6b82bA149CFA113f8f48d5E3b1F78e933e16DfD,1000000000000000000",
            ],
        ]
        self.assertEqual(safe_operator.get_transaction_history(), expected)

    def test_data_decoded_to_text(self):
        safe_operator = self.setup_operator(
            number_owners=1, mode=SafeOperatorMode.TX_SERVICE
        )
        decoded_data_text = safe_operator.safe_tx_service.data_decoded_to_text(
            data_decoded_mock
        )
        self.assertIn(
            "- changeMasterCopy: 0x34CfAC646f301356fAa8B21e94227e3583Fe3F5F",
            decoded_data_text,
        )
        self.assertIn(
            "- setFallbackHandler: 0xd5D82B6aDDc9027B22dCA772Aa68D5d74cdBdF44",
            decoded_data_text,
        )

    def test_drain(self):
        # TODO Drain is a complex to mock
        pass


if __name__ == "__main__":
    unittest.main()



from prompt_toolkit.formatted_text import HTML

SAFE_ARGUMENT_COLOR = "em"
SAFE_EMPTY_ARGUMENT_COLOR = "ansimagenta"


safe_commands_arguments = {
    "add_delegate": "<address> <label> <signer-address>",
    "add_owner": "<address> [--threshold <int>]",
    "approve_hash": "<safe-tx-hash> <address>",
    "balances": "(read-only)",
    "batch-txs": "<safe-nonce> <safe-tx-hash> [ <safe-tx-hash> ... ]",
    "change_fallback_handler": "<address>",
    "change_guard": "<address>",
    "change_master_copy": "<address>",
    "change_threshold": "<address>",
    "disable_module": "<address>",
    "enable_module": "<address>",
    "execute-tx": "<safe-tx-hash>",
    "get_delegates": "(read-only)",
    "get_nonce": "(read-only)",
    "get_owners": "(read-only)",
    "get_threshold": "(read-only)",
    "history": "(read-only)",
    "info": "(read-only)",
    "load_cli_owners": "<account-private-key> [<account-private-key>...]",
    "load_ledger_cli_owners": "[--legacy-accounts] [--derivation-path <str>]",
    "load_trezor_cli_owners": "[--legacy-accounts] [--derivation-path <str>]",
    "load_cli_owners_from_words": "<word_1> <word_2> ... <word_12>",
    "refresh": "",
    "remove_delegate": "<address> <signer-address>",
    "remove_owner": "<address> [--threshold <int>]",
    "remove_proposed_transaction": "safe-tx-hash",
    "send_custom": "<address> <value-wei> <data> [--delegate] [--safe-nonce <int>]",
    "send_erc20": "<address> <token-address> <value-wei> [--safe-nonce <int>]",
    "send_erc721": "<address> <token-address> <token-id> [--safe-nonce <int>]",
    "send_ether": "<address> <value-wei> [--safe-nonce <int>]",
    "show_cli_owners": "(read-only)",
    "sign_message": "[--eip191_message <str>] [--eip712_path <file-path>]",
    "sign-tx": "<safe-tx-hash>",
    "unload_cli_owners": "<address> [<address>...]",
    "update": "",
    "update_version_to_l2": "<address>",
    "blockchain": "",
    "tx-service": "",
    "drain": "<address>",
    "fs_attestation_service_enrolment": "",
    "fs_attestation_service_auth": "<safe_cli_owner> <safe_cli_owner_private_key>",
    "fs_attestation_service_deploy_guard_contract": "",
    "fs_attestation_service_set_constraints": "<signer_wallet_address:ip_address_range> [<signer_wallet_address:ip_address_range>...] <multi_sig_tx_expiry_duration> <risk_score_threshold>",
    "fs_attestation_settings": "<contact_email> <slack_web_hook_url>"
}

safe_commands = list(safe_commands_arguments.keys())

safe_color_arguments = {
    "(read-only)": SAFE_ARGUMENT_COLOR,
    "<account-private-key>": SAFE_ARGUMENT_COLOR,
    "<address>": SAFE_ARGUMENT_COLOR,
    "<hex-str>": SAFE_ARGUMENT_COLOR,
    "<integer>": SAFE_ARGUMENT_COLOR,
    "<safe-tx-hash>": SAFE_ARGUMENT_COLOR,
    "<token-address>": SAFE_ARGUMENT_COLOR,
    "<token-id>": SAFE_ARGUMENT_COLOR,
    "<value-wei>": SAFE_ARGUMENT_COLOR,
}

meta = {
    "approve_hash": HTML(
        "<b>approve_hash</b> will approve a safe-tx-hash for the provided sender address. "
        "Sender private key must be loaded first"
    ),
    "balances": HTML(
        "<b>balances</b> will return the balance of Ether and ERC20 tokens of the Safe "
        "(if tx service available for the network)"
    ),
    "history": HTML(
        "<b>history</b> will return information of last transactions for the Safe "
        "(if tx service available for the network)"
    ),
    "batch-txs": HTML(
        "<b>batch-txs</b> will take pending or executed transactions by safe tx hash and will create a new"
        "transaction using the provided safe nonce"
    ),
    "execute-tx": HTML(
        "Take a pending transaction from Safe Tx Service and execute it using a loaded sender"
    ),
    "sign-tx": HTML(
        "<b>sign-tx</b> will sign the provided safeTxHash using the owners loaded on the CLI"
    ),
    "sign_message": HTML(
        "<b>sign_message</b> sign the provided string message provided by standard input or the EIP712 provided by file"
    ),
    "info": HTML(
        "<b>info</b> will return all the information available for a Safe, with Gnosis Tx Service and "
        "Etherscan links if the network is supported"
    ),
    "show_cli_owners": HTML(
        "Command <b>show_cli_owners</b> will return a list of loaded <u>&lt;address&gt;</u> "
        "account owners."
    ),
    "get_owners": HTML(
        "Command <b>get_owners</b> will return a list of check-summed <u>&lt;address&gt;</u> "
        "account owners."
    ),
    "get_delegates": HTML(
        "Command <b>get_delegates</b> will return information about the current delegates."
    ),
    "change_owner": HTML(
        "Command <b>change_owner</b> will change an old account <u>&lt;address&gt;</u> for the new "
        "check-summed <u>&lt;address&gt;</u> account."
    ),
    "add_owner": HTML(
        "Command <b>add_owner</b> will add a check-summed <u>&lt;address&gt;</u> owner account."
    ),
    "remove_owner": HTML(
        "Command <b>remove_owner</b> will remove an old account <u>&lt;address&gt;</u> from the "
        "current loaded safe."
    ),
    "add_delegate": HTML(
        "Command <b>add_delegate</b> will add a check-summed <u>&lt;address&gt;</u> delegate account."
    ),
    "remove_delegate": HTML(
        "Command <b>remove_delegate</b> will remove a delegate <u>&lt;address&gt;</u> from the "
        "current loaded safe."
    ),
    "remove_proposed_transaction": HTML(
        "Command <b>remove_proposed_transaction</b> will remove proposed not executed transaction for "
        "provided<u>&lt;safe-tx-hash&gt;</u>"
    ),
    "enable_module": HTML(
        "Command <b>enable_module</b> will enable a check-summed <u>&lt;address&gt;</u> module."
    ),
    "disable_module": HTML(
        "Command <b>disable_module</b> will disable a check-summed <u>&lt;address&gt;</u> module."
    ),
    "get_threshold": HTML(
        "Command <b>get_threshold</b> will return the threshold <u>&lt;value&gt;</u> for"
        " the current loaded safe."
    ),
    "get_nonce": HTML(
        "Command <b>get_nonce</b> will return the nonce <u>&lt;value&gt;</u> for "
        "the current loaded safe."
    ),
    "change_threshold": HTML(
        "Command <b>change_threshold</b> will change the current threshold <u>&lt;integer&gt;</u> "
        "value for the loaded safe."
    ),
    "send_custom": HTML(
        "Command <b>send_custom</b> will try to send a custom tx to a check-summed account. Set value "
        "to 0 if you don't want to send ether. <b>--delegate</b> can be added to send a DELEGATECALL"
    ),
    "send_ether": HTML(
        "Command <b>send_ether</b> will try to send Wei <u>&lt;value&gt;</u> to a check-summed account"
        " <u>&lt;address&gt;</u> if enough funds are found, withing the current loaded safe."
    ),
    "send_erc20": HTML(
        "Command <b>send_erc20</b> will try to send a Token <u>&lt;value&gt;</u> from a check-summed "
        "<u>&lt;token-address&gt;</u>, to a check-summed account <u>&lt;address&gt;</u> if enough funds"
        " are found, withing the current loaded safe."
    ),
    "send_erc721": HTML(
        "Command <b>send_erc721</b> will try to send a ERC 721 Token <u>&lt;value&gt;</u>"
        "from a check-summed <u>&lt;token-address&gt;</u>, to a check-summed account "
        "<u>&lt;address&gt;</u>."
    ),
    "unload_cli_owners": HTML(
        "Command <b>unload_cli_owners</b> will unload a check-summed <u>&lt;address&gt;</u> "
        "from the current loaded account owners."
    ),
    "load_cli_owners": HTML(
        "Command <b>load_cli_owners</b> will try to load a new owner via "
        "<u>&lt;account-private-key&gt;</u>."
    ),
    "load_ledger_cli_owners": HTML(
        "Command <b>load_ledger_cli_owners</b> show a list of Ledger hardware wallet addresses to choose between them "
    ),
    "load_trezor_cli_owners": HTML(
        "Command <b>load_trezor_cli_owners</b> show a list of Trezor hardware wallet addresses to choose between them "
    ),
    "load_cli_owners_from_words": HTML(
        "Command <b>load_cli_owners_from_words</b> will try to load owners via"
        "<u>seed_words</u>. Only relevant accounts(owners) will be loaded"
    ),
    "refresh": HTML(
        "Command <b>refresh</b> will refresh the information for the current loaded safe."
    ),
    "change_master_copy": HTML(
        "Command <b>change_master_copy</b> will change the current MasterCopy of the "
        "Safe Contract <b>[DO NOT CALL THIS FUNCTION, UNLESS YOU KNOW WHAT YOU ARE DOING. "
        "ALL YOUR FUNDS COULD BE LOST]</b>."
    ),
    "change_fallback_handler": HTML(
        "Command <b>change_fallback_handler</b> will change the current "
        "fallbackHandler for Safes with version >= 1.1.0 "
        "<b>[DO NOT CALL THIS FUNCTION, UNLESS YOU KNOW WHAT YOU ARE DOING. "
        "ALL YOUR FUNDS COULD BE LOST]</b>."
    ),
    "change_guard": HTML(
        "Command <b>change_guard</b> will change the current "
        "guard for Safes with version >= 1.3.0 "
        "<b>[DO NOT CALL THIS FUNCTION, UNLESS YOU KNOW WHAT YOU ARE DOING. "
        "ALL YOUR FUNDS COULD BE LOST]</b>."
    ),
    "update": HTML(
        "Command <b>update</b> will upgrade the Safe master copy to the latest version"
    ),
    "update_version_to_l2": HTML(
        "Updates a v1.1.1/v1.3.0/v1.4.1 non L2 Safe to a L2 Safe supported by Safe Wallet UI. "
        "The migration contract address needs to be provided. Nonce for the Safe must be 0."
    ),
    "blockchain": HTML(
        "<b>blockchain</b> sets the default mode for tx service. Transactions will be "
        "sent to blockchain"
    ),
    "tx-service": HTML(
        "<b>tx-service</b> enables tx-service integration. Transactions will be sent to the tx-service "
        "instead of blockchain, so they will show up on the interface"
    ),
    "drain": HTML(
        "Command <b>drain</b> will try to send all assets ether and ERC20 to a check-summed account"
    ),
    "fs_attestation_service_auth": HTML(
        "Performs authentication into FailSafe Attestation Service using one of the owners wallet address and private key"
    ),
    "fs_attestation_service_enrolment": HTML(
        "Will enroll the current safe wallet address into FailSafe Attestation Service"
    ),
    "fs_attestation_service_deploy_guard_contract": HTML(
        "Will deploy a FailSafe Attestation Service guard contract for the current safe wallet address"
    ),
    "fs_attestation_service_set_constraints": HTML(
        "Will set the ip address range, multi-sig transaction expiry duration and risk score threshold constraints for the current safe wallet address"
    ),
    "fs_attestation_settings": HTML(
        "Will update the additional settings FailSafe Attestation Service settings (contact emails and slack integrations, etc) for the current safe wallet address"
    ),
}


import json
import unittest
from functools import lru_cache
from unittest import mock
from unittest.mock import MagicMock, PropertyMock

from eth_account import Account
from eth_typing import ChecksumAddress
from ledgerblue.Dongle import Dongle
from ledgereth.objects import LedgerAccount
from web3 import Web3
from web3.types import Wei

from gnosis.eth import EthereumClient
from gnosis.eth.eip712 import eip712_encode
from gnosis.safe import Safe
from gnosis.safe.multi_send import MultiSend

from safe_cli.contracts import safe_to_l2_migration
from safe_cli.operators.exceptions import (
    AccountNotLoadedException,
    ExistingOwnerException,
    FallbackHandlerNotSupportedException,
    GuardNotSupportedException,
    HardwareWalletException,
    HashAlreadyApproved,
    InvalidFallbackHandlerException,
    InvalidGuardException,
    InvalidMasterCopyException,
    NonExistingOwnerException,
    NotEnoughEtherToSend,
    NotEnoughSignatures,
    SafeVersionNotSupportedException,
    SameFallbackHandlerException,
    SameGuardException,
    SameMasterCopyException,
    SenderRequiredException,
)
from safe_cli.operators.safe_operator import SafeOperator
from safe_cli.utils import get_erc_20_list
from tests.utils import generate_transfers_erc20

from .safe_cli_test_case_mixin import SafeCliTestCaseMixin


class TestSafeOperator(SafeCliTestCaseMixin, unittest.TestCase):
    @lru_cache(maxsize=None)
    def _deploy_l2_migration_contract(self) -> ChecksumAddress:
        # Deploy L2 migration contract
        safe_to_l2_migration_contract = self.w3.eth.contract(
            abi=safe_to_l2_migration["abi"], bytecode=safe_to_l2_migration["bytecode"]
        )
        tx_hash = safe_to_l2_migration_contract.constructor().transact(
            {"from": self.ethereum_test_account.address}
        )
        tx_receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash)
        return tx_receipt["contractAddress"]

    def test_setup_operator(self):
        for number_owners in range(1, 4):
            safe_operator = self.setup_operator(number_owners=number_owners)
            self.assertEqual(len(safe_operator.accounts), number_owners)
            safe = Safe(safe_operator.address, self.ethereum_client)
            self.assertEqual(len(safe.retrieve_owners()), number_owners)

    @mock.patch(
        "gnosis.safe.Safe.contract", new_callable=mock.PropertyMock, return_value=None
    )
    def test_load_cli_owner(self, get_contract_mock: MagicMock):
        random_address = Account.create().address
        safe_operator = SafeOperator(random_address, self.ethereum_node_url)
        random_accounts = [Account.create() for _ in range(3)]
        random_accounts_keys = [account.key.hex() for account in random_accounts]
        self.assertFalse(safe_operator.accounts)
        safe_operator.load_cli_owners(random_accounts_keys)
        self.assertEqual(len(safe_operator.accounts), len(random_accounts_keys))
        # Test accounts are not duplicated
        safe_operator.load_cli_owners(random_accounts_keys)
        self.assertEqual(len(safe_operator.accounts), len(random_accounts_keys))
        # Test invalid accounts don't make the method break
        safe_operator.load_cli_owners(["aloha", Account.create().key.hex(), "bye"])
        self.assertEqual(len(safe_operator.accounts), len(random_accounts_keys) + 1)
        # TODO Test setting default sender, mock getBalance

        # Test unload cli owner
        safe_operator.default_sender = random_accounts[0]
        number_of_accounts = len(safe_operator.accounts)

        safe_operator.unload_cli_owners(["aloha", random_accounts[0].address, "bye"])
        self.assertEqual(len(safe_operator.accounts), number_of_accounts - 1)
        self.assertFalse(safe_operator.default_sender)

    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.init_dongle",
        return_value=Dongle(),
    )
    @mock.patch("safe_cli.operators.hw_wallets.ledger_wallet.get_account_by_path")
    def test_load_ledger_cli_owner(
        self, mock_get_account_by_path: MagicMock, mock_init_dongle: MagicMock
    ):
        owner_address = Account.create().address
        safe_address = self.deploy_test_safe(owners=[owner_address]).address
        safe_operator = SafeOperator(safe_address, self.ethereum_node_url)
        safe_operator.hw_wallet_manager.get_accounts = MagicMock(return_value=[])
        safe_operator.load_ledger_cli_owners()
        self.assertEqual(len(safe_operator.hw_wallet_manager.wallets), 0)
        random_address = Account.create().address
        other_random_address = Account.create().address
        safe_operator.hw_wallet_manager.get_accounts.return_value = [
            (random_address, "44'/60'/0'/0/0"),
            (other_random_address, "44'/60'/0'/0/1"),
        ]

        mock_get_account_by_path.return_value = LedgerAccount(
            "44'/60'/0'/0/0", random_address
        )
        safe_operator.load_ledger_cli_owners()
        self.assertEqual(len(safe_operator.hw_wallet_manager.wallets), 1)
        # Wallet without funds shouldn't be added as sender
        self.assertIsNone(safe_operator.hw_wallet_manager.sender)
        self.assertEqual(
            safe_operator.hw_wallet_manager.wallets.pop().address, random_address
        )

        # Only accept ethereum derivation paths
        with self.assertRaises(HardwareWalletException):
            safe_operator.load_ledger_cli_owners(derivation_path="44'/137'/0'/0/1")
        mock_get_account_by_path.return_value = LedgerAccount(
            "44'/60'/0'/0/0", owner_address
        )
        safe_operator.load_ledger_cli_owners(derivation_path="44'/60'/0'/0/0")
        self.assertEqual(len(safe_operator.hw_wallet_manager.wallets), 1)
        self.assertEqual(
            safe_operator.hw_wallet_manager.wallets.pop().address, owner_address
        )

        # Should be loaded as sender
        ledger_random_address = Account.create().address
        # Send funds to define it as sender
        self.send_ether(ledger_random_address, Wei(1000))
        mock_get_account_by_path.return_value = LedgerAccount(
            "44'/60'/0'/0/1", ledger_random_address
        )
        safe_operator.load_ledger_cli_owners(derivation_path="44'/60'/0'/0/1")
        self.assertEqual(len(safe_operator.hw_wallet_manager.wallets), 1)
        self.assertEqual(
            safe_operator.hw_wallet_manager.sender.address, ledger_random_address
        )

        # test unload ledger owner
        safe_operator.unload_cli_owners([ledger_random_address])
        self.assertEqual(len(safe_operator.hw_wallet_manager.wallets), 0)
        self.assertIsNone(safe_operator.hw_wallet_manager.sender)

    def test_approve_hash(self):
        safe_address = self.deploy_test_safe(
            owners=[self.ethereum_test_account.address]
        ).address
        safe_operator = SafeOperator(safe_address, self.ethereum_node_url)
        safe_tx_hash = Web3.keccak(text="random-test")
        random_account = Account.create()
        with self.assertRaises(AccountNotLoadedException):
            safe_operator.approve_hash(safe_tx_hash, random_account.address)

        with self.assertRaises(NonExistingOwnerException):
            safe_operator.accounts.add(random_account)
            safe_operator.approve_hash(safe_tx_hash, random_account.address)

        safe_operator.accounts.add(self.ethereum_test_account)
        safe_operator.default_sender = self.ethereum_test_account
        self.assertTrue(
            safe_operator.approve_hash(safe_tx_hash, self.ethereum_test_account.address)
        )
        with self.assertRaises(HashAlreadyApproved):
            safe_operator.approve_hash(safe_tx_hash, self.ethereum_test_account.address)

    @mock.patch("safe_cli.safe_addresses._get_valid_contract")
    def test_sign_message(self, mock_sign_message_lib_address):
        safe_address = self.deploy_test_safe(
            owners=[self.ethereum_test_account.address]
        ).address
        safe_operator = SafeOperator(safe_address, self.ethereum_node_url)
        mock_sign_message_lib_address.return_value = self.deploy_sign_message_lib()
        message = "Safe2024"
        safe_operator.accounts.add(self.ethereum_test_account)
        safe_operator.default_sender = self.ethereum_test_account
        message_hash = safe_operator.safe.get_message_hash(bytes(message, "utf-8"))
        safe_operator.sign_message(eip191_message=message)
        self.assertTrue(safe_operator.safe.retrieve_is_message_signed(message_hash))
        eip712_path = "tests/mocks/mock_eip712.json"
        message = json.load(open(eip712_path, "r"))
        message_hash = safe_operator.safe.get_message_hash(
            b"".join(eip712_encode(message))
        )
        self.assertFalse(safe_operator.safe.retrieve_is_message_signed(message_hash))
        safe_operator.sign_message(eip712_message_path=eip712_path)
        self.assertTrue(safe_operator.safe.retrieve_is_message_signed(message_hash))

    def test_add_owner(self):
        safe_address = self.deploy_test_safe(
            owners=[self.ethereum_test_account.address]
        ).address
        safe_operator = SafeOperator(safe_address, self.ethereum_node_url)
        with self.assertRaises(ExistingOwnerException):
            safe_operator.add_owner(self.ethereum_test_account.address)

        new_owner = Account.create().address
        with self.assertRaises(NotEnoughSignatures):
            safe_operator.add_owner(new_owner)

        safe_operator.accounts.add(self.ethereum_test_account)

        with self.assertRaises(SenderRequiredException):
            safe_operator.add_owner(new_owner)

        safe_operator.default_sender = self.ethereum_test_account

        safe = Safe(safe_address, self.ethereum_client)
        self.assertTrue(safe_operator.add_owner(new_owner))
        self.assertIn(self.ethereum_test_account, safe_operator.accounts)
        self.assertIn(new_owner, safe.retrieve_owners())

    def test_remove_owner(self):
        safe_address = self.deploy_test_safe(
            owners=[self.ethereum_test_account.address]
        ).address
        safe_operator = SafeOperator(safe_address, self.ethereum_node_url)
        random_address = Account.create().address
        with self.assertRaises(NonExistingOwnerException):
            safe_operator.remove_owner(random_address)

        safe_operator.load_cli_owners([self.ethereum_test_account.key.hex()])
        new_owner = Account.create().address
        safe = Safe(safe_address, self.ethereum_client)
        self.assertTrue(safe_operator.add_owner(new_owner))
        self.assertIn(new_owner, safe.retrieve_owners())

        self.assertTrue(safe_operator.remove_owner(new_owner))
        self.assertNotIn(new_owner, safe_operator.accounts)
        self.assertNotIn(new_owner, safe.retrieve_owners())

    def test_change_fallback_handler(self):
        safe_operator = self.setup_operator()
        safe = Safe(safe_operator.address, self.ethereum_client)
        current_fallback_handler = safe.retrieve_fallback_handler()
        with self.assertRaises(SameFallbackHandlerException):
            safe_operator.change_fallback_handler(current_fallback_handler)

        new_fallback_handler = Account.create().address
        with self.assertRaises(InvalidFallbackHandlerException):
            # Contract does not exist
            self.assertTrue(safe_operator.change_fallback_handler(new_fallback_handler))

        with mock.patch.object(
            EthereumClient, "is_contract", autospec=True, return_value=True
        ):
            self.assertTrue(safe_operator.change_fallback_handler(new_fallback_handler))
        self.assertEqual(
            safe_operator.safe_cli_info.fallback_handler, new_fallback_handler
        )
        self.assertEqual(safe.retrieve_fallback_handler(), new_fallback_handler)

        # Safes < 1.1.0 don't support the fallback handler
        safe_operator_v1_0_0 = self.setup_operator(version="1.0.0")
        with self.assertRaises(FallbackHandlerNotSupportedException):
            safe_operator_v1_0_0.change_fallback_handler(Account.create().address)

    def test_change_guard(self):
        safe_operator = self.setup_operator(version="1.1.1")
        with self.assertRaises(GuardNotSupportedException):
            safe_operator.change_guard(Account.create().address)

        safe_operator = self.setup_operator(version="1.4.1")
        safe = Safe(safe_operator.address, self.ethereum_client)
        current_guard = safe.retrieve_guard()
        with self.assertRaises(SameGuardException):
            safe_operator.change_guard(current_guard)

        not_valid_guard = Account.create().address
        with self.assertRaises(InvalidGuardException):  # Contract does not exist
            self.assertTrue(safe_operator.change_guard(not_valid_guard))

        new_guard = self.deploy_example_guard()
        self.assertTrue(safe_operator.change_guard(new_guard))
        self.assertEqual(safe_operator.safe_cli_info.guard, new_guard)
        self.assertEqual(safe.retrieve_guard(), new_guard)

    def test_change_master_copy(self):
        safe_operator = self.setup_operator(version="1.3.0")
        with self.assertRaises(SafeVersionNotSupportedException):
            safe_operator.change_master_copy(self.safe_contract_V1_4_1.address)

        safe_operator = self.setup_operator(version="1.1.1")
        safe = Safe(safe_operator.address, self.ethereum_client)
        current_master_copy = safe.retrieve_master_copy_address()
        self.assertEqual(current_master_copy, self.safe_contract_V1_1_1.address)
        with self.assertRaises(SameMasterCopyException):
            safe_operator.change_master_copy(current_master_copy)

        random_address = Account.create().address
        with self.assertRaises(InvalidMasterCopyException):
            safe_operator.change_master_copy(random_address)

        new_master_copy = self.safe_contract_V1_3_0.address
        self.assertTrue(safe_operator.change_master_copy(new_master_copy))
        self.assertEqual(safe_operator.safe_cli_info.master_copy, new_master_copy)
        self.assertEqual(safe.retrieve_master_copy_address(), new_master_copy)

    def test_send_ether(self):
        safe_operator = self.setup_operator()
        random_address = Account.create().address
        value = 123
        with self.assertRaises(NotEnoughEtherToSend):
            safe_operator.send_ether(random_address, value)
        self.ethereum_client.send_eth_to(
            self.ethereum_test_account.key,
            safe_operator.address,
            self.w3.eth.gas_price,
            value * 2,
            gas=50000,
        )
        self.assertEqual(
            self.ethereum_client.get_balance(safe_operator.address), value * 2
        )
        self.assertTrue(safe_operator.send_ether(random_address, value))
        self.assertEqual(self.ethereum_client.get_balance(random_address), value)

    @mock.patch.object(
        SafeOperator, "last_default_fallback_handler_address", new_callable=PropertyMock
    )
    @mock.patch.object(
        SafeOperator, "last_safe_contract_address", new_callable=PropertyMock
    )
    def test_update_version(
        self,
        last_safe_contract_address_mock: PropertyMock,
        last_default_fallback_handler_address: PropertyMock,
    ):
        last_safe_contract_address_mock.return_value = self.safe_contract_V1_4_1.address
        last_default_fallback_handler_address.return_value = (
            self.compatibility_fallback_handler.address
        )

        safe_operator_v130 = self.setup_operator(version="1.3.0")
        with self.assertRaises(SafeVersionNotSupportedException):
            safe_operator_v130.update_version()

        safe_operator_v111 = self.setup_operator(version="1.1.1")
        with mock.patch.object(
            MultiSend,
            "MULTISEND_CALL_ONLY_ADDRESSES",
            [self.multi_send_contract.address],
        ):
            safe_operator_v111.update_version()

        self.assertEqual(
            safe_operator_v111.safe.retrieve_master_copy_address(),
            last_safe_contract_address_mock.return_value,
        )
        self.assertEqual(
            safe_operator_v111.safe.retrieve_fallback_handler(),
            last_default_fallback_handler_address.return_value,
        )

    def test_update_to_l2_v111(self):
        migration_contract_address = self._deploy_l2_migration_contract()
        safe_operator_v111 = self.setup_operator(version="1.1.1")

        with mock.patch.dict(
            "safe_cli.operators.safe_operator.safe_deployments",
            {
                "1.3.0": {
                    "GnosisSafeL2": {"1337": self.safe_contract_V1_3_0.address},
                    "CompatibilityFallbackHandler": {
                        "1337": self.compatibility_fallback_handler.address
                    },
                }
            },
        ):
            self.assertEqual(safe_operator_v111.safe.retrieve_version(), "1.1.1")
            safe_operator_v111.update_version_to_l2(migration_contract_address)
            self.assertEqual(
                safe_operator_v111.safe.retrieve_master_copy_address(),
                self.safe_contract_V1_3_0.address,
            )
            self.assertEqual(
                safe_operator_v111.safe.retrieve_fallback_handler(),
                self.compatibility_fallback_handler.address,
            )

    def test_update_to_l2_v130(self):
        migration_contract_address = self._deploy_l2_migration_contract()
        safe_operator_v130 = self.setup_operator(version="1.3.0")

        # For testing v1.3.0 non L2 to L2 update we need 1.3.0 deployed in a different address
        # L2 Migration Contract only checks version but cannot tell apart L2 from not L2
        tx_hash = self.safe_contract_V1_3_0.constructor().transact(
            {"from": self.ethereum_test_account.address}
        )
        tx_receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash)
        safe_contract_l2_130_address = tx_receipt["contractAddress"]
        with mock.patch.dict(
            "safe_cli.operators.safe_operator.safe_deployments",
            {
                "1.3.0": {
                    "GnosisSafeL2": {"1337": safe_contract_l2_130_address},
                }
            },
        ):
            self.assertEqual(safe_operator_v130.safe.retrieve_version(), "1.3.0")
            self.assertEqual(
                safe_operator_v130.safe.retrieve_master_copy_address(),
                self.safe_contract_V1_3_0.address,
            )
            previous_fallback_handler = (
                safe_operator_v130.safe.retrieve_fallback_handler()
            )
            safe_operator_v130.update_version_to_l2(migration_contract_address)
            self.assertEqual(
                safe_operator_v130.safe.retrieve_master_copy_address(),
                safe_contract_l2_130_address,
            )
            self.assertEqual(
                safe_operator_v130.safe.retrieve_fallback_handler(),
                previous_fallback_handler,
            )

    def test_drain(self):
        safe_operator = self.setup_operator()
        account = Account.create()
        value = self.w3.to_wei(10.5, "ether")

        with self.assertRaises(NotEnoughEtherToSend):
            safe_operator.send_ether(account.address, value)

        self.ethereum_client.send_eth_to(
            self.ethereum_test_account.key,
            safe_operator.address,
            self.w3.eth.gas_price,
            value,
            gas=50000,
        )
        safe_operator.send_ether(account.address, value - 5)
        # Deploying 3 contracts and sending 3 transactions of 1 ERC20 per contract
        num_transactions = 3
        num_contracts_erc20 = 3
        generate_transfers_erc20(
            self.w3, safe_operator, account, num_contracts_erc20, num_transactions
        )
        # Deploying multisend contract on ganache
        _, _, contract_address = MultiSend.deploy_contract(
            self.ethereum_client, account
        )
        with mock.patch.object(
            MultiSend, "MULTISEND_CALL_ONLY_ADDRESSES", [contract_address]
        ):
            # Getting events filtered by Transfer
            last = safe_operator.ethereum_client.get_block("latest")["number"]
            token_address = get_erc_20_list(
                safe_operator.ethereum_client, safe_operator.address, 1, last
            )
            self.assertEqual(len(token_address), num_contracts_erc20)
            for token in token_address:
                result = self.ethereum_client.erc20.get_balance(
                    safe_operator.address, token
                )
                self.assertEqual(result, num_transactions)
            # Draining the account to a new account
            safe_operator.drain(account.address)
            # Checking that the account is empty
            for token in token_address:
                self.assertEqual(
                    self.ethereum_client.erc20.get_balance(
                        safe_operator.address, token
                    ),
                    0,
                )
            self.assertEqual(
                safe_operator.ethereum_client.get_balance(safe_operator.address), 0
            )


if __name__ == "__main__":
    unittest.main()



import argparse
import unittest
from unittest import mock
from unittest.mock import MagicMock

from eth_account import Account

from gnosis.safe import Safe

from safe_cli.safe_creator import main

from .safe_cli_test_case_mixin import SafeCliTestCaseMixin


class TestSafeCreator(SafeCliTestCaseMixin, unittest.TestCase):
    @mock.patch(
        "argparse.ArgumentParser.parse_args",
    )
    def test_main(self, mock_parse_args: MagicMock):
        owner_account = self.get_ethereum_test_account()
        owners = [Account.create().address]
        threshold = 1
        mock_parse_args.return_value = argparse.Namespace(
            owners=owners,
            threshold=threshold,
            salt_nonce=4815,
            node_url=self.ethereum_node_url,
            private_key=owner_account.key.hex(),
            safe_contract=self.safe_contract.address,
            proxy_factory=self.proxy_factory_contract.address,
            callback_handler=self.compatibility_fallback_handler.address,
        )

        safe_address = main().contract_address
        safe = Safe(safe_address, self.ethereum_client)
        safe_info = safe.retrieve_all_info()
        self.assertEqual(safe_info.owners, owners)
        self.assertEqual(safe_info.threshold, 1)
        self.assertEqual(safe_info.master_copy, self.safe_contract.address)
        self.assertEqual(
            safe_info.fallback_handler, self.compatibility_fallback_handler.address
        )

        # If contract is deployed script should exit with code 1
        with self.assertRaisesRegex(SystemExit, "1"):
            main()


if __name__ == "__main__":
    unittest.main()



#!/bin/env python3
import argparse
import os
import sys
from typing import Optional

from art import text2art, tprint
from eth_typing import ChecksumAddress
from prompt_toolkit import HTML, PromptSession, print_formatted_text
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.history import FileHistory
from prompt_toolkit.lexers import PygmentsLexer

from safe_cli.argparse_validators import check_ethereum_address
from safe_cli.operators import (
    SafeOperator,
    SafeServiceNotAvailable,
    SafeTxServiceOperator,
    FailSafeAttestationServiceOperator
)
from safe_cli.prompt_parser import PromptParser
from safe_cli.safe_completer import SafeCompleter
from safe_cli.safe_lexer import SafeLexer
from safe_cli.utils import get_safe_from_owner

from .version import version


class SafeCli:
    def __init__(self, safe_address: ChecksumAddress, node_url: str, safe_tx_service_base_url: str, fs_attestation_service_base_url: str, history: bool):
        """
        :param safe_address: Safe address
        :param node_url: Ethereum RPC url
        :param history: If `True` keep command history, otherwise history is not kept after closing the CLI
        """
        self.safe_address = safe_address
        self.node_url = node_url
        self.fs_attestation_service_base_url = fs_attestation_service_base_url
        if history:
            self.session = PromptSession(
                history=FileHistory(os.path.join(sys.path[0], ".history"))
            )
        else:
            self.session = PromptSession()
        self.safe_operator = SafeOperator(safe_address, node_url)
        self.fs_attestation_service_operator = FailSafeAttestationServiceOperator(safe_address, node_url, safe_tx_service_base_url, fs_attestation_service_base_url)
        self.prompt_parser = PromptParser(self.safe_operator, self.fs_attestation_service_operator)

    def print_startup_info(self):
        print_formatted_text(tprint("FailSafe Attestation"))  # Print fancy text
        print_formatted_text(text2art("Safe CLI"))  # Print fancy text
        print_formatted_text(HTML(f"<b><ansigreen>Version {version}</ansigreen></b>"))
        print_formatted_text(
            HTML("<b><ansigreen>Loading Safe information...</ansigreen></b>")
        )
        self.safe_operator.print_info()

    def get_prompt_text(self):
        mode: Optional[str] = "blockchain"
        if isinstance(self.prompt_parser.safe_operator, SafeTxServiceOperator):
            mode = "tx-service"

        return HTML(
            f"<bold><ansiblue>{mode} > {self.safe_address}</ansiblue><ansired> > </ansired></bold>"
        )

    def get_bottom_toolbar(self):
        return HTML(
            f'<b><style fg="ansiyellow">network={self.safe_operator.network.name} '
            f"{self.safe_operator.safe_cli_info}</style></b>"
        )

    def parse_operator_mode(self, command: str) -> Optional[SafeOperator]:
        """
        Parse operator mode to switch between blockchain (default) and tx-service
        :param command:
        :return: SafeOperator if detected
        """
        split_command = command.split()
        try:
            if (split_command[0]) == "tx-service":
                print_formatted_text(
                    HTML("<b><ansigreen>Sending txs to tx service</ansigreen></b>")
                )
                return SafeTxServiceOperator(self.safe_address, self.node_url)
            elif split_command[0] == "blockchain":
                print_formatted_text(
                    HTML("<b><ansigreen>Sending txs to blockchain</ansigreen></b>")
                )
                return self.safe_operator
        except SafeServiceNotAvailable:
            print_formatted_text(
                HTML("<b><ansired>Mode not supported on this network</ansired></b>")
            )

    def get_command(self) -> str:
        return self.session.prompt(
            self.get_prompt_text,
            auto_suggest=AutoSuggestFromHistory(),
            bottom_toolbar=self.get_bottom_toolbar,
            lexer=PygmentsLexer(SafeLexer),
            completer=SafeCompleter(),
        )

    def loop(self):
        while True:
            try:
                command = self.get_command()
                if not command.strip():
                    continue

                new_operator = self.parse_operator_mode(command)
                if new_operator:
                    self.prompt_parser = PromptParser(new_operator)
                    new_operator.refresh_safe_cli_info()  # ClI info needs to be initialized
                else:
                    self.prompt_parser.process_command(command)
            except EOFError:
                break
            except KeyboardInterrupt:
                continue
            except (argparse.ArgumentError, argparse.ArgumentTypeError, SystemExit):
                pass


def build_safe_cli() -> Optional[SafeCli]:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "address",
        help="The address of the Safe, or an owner address if --get-safes-from-owner is specified.",
        type=check_ethereum_address,
    )
    parser.add_argument("node_url", help="Ethereum node url")
    parser.add_argument("safe_tx_service_base_url", help="Safe Transaction Service Base URL")
    parser.add_argument("fs_attestation_service_base_url", help="FailSafe Attestation Service API Base URL")
    parser.add_argument(
        "--history",
        action="store_true",
        help="Enable history. By default it's disabled due to security reasons",
        default=False,
    )
    parser.add_argument(
        "--get-safes-from-owner",
        action="store_true",
        help="Indicates that address is an owner (Safe Transaction Service is required for this feature)",
        default=False,
    )

    args = parser.parse_args()
    if args.get_safes_from_owner:
        if (
            safe_address := get_safe_from_owner(args.address, args.node_url)
        ) is not None:
            return SafeCli(safe_address, args.node_url, args.safe_tx_service_base_url, args.fs_attestation_service_base_url, args.history)
    else:
        return SafeCli(args.address, args.node_url, args.safe_tx_service_base_url, args.fs_attestation_service_base_url, args.history)


def main(*args, **kwargs):
    safe_cli = build_safe_cli()
    safe_cli.print_startup_info()
    safe_cli.loop()


if __name__ == "__main__":
    main()


import re

from pygments.lexers.shell import BashLexer
from pygments.token import Keyword, Name, Text, Token


class SafeLexer(BashLexer):
    name = "SafeLexer"
    aliases = ["safe_lexer"]

    ADDRESS = r"^0x[aA-zZ,0-9]{40}$|^0x[aA-zZ,0-9]{62}$"
    EXTRA_KEYWORDS = {
        "refresh",
        "get_nonce",
        "get_owners",
        "get_threshold",
        "get_delegates",
        "show_cli_owners",
        "load_cli_owners_from_words",
        "load_cli_owners",
        "unload_cli_owners",
        "approve_hash",
        "sign_message",
        "add_owner",
        "change_threshold",
        "change_fallback_handler",
        "change_guard",
        "remove_owner",
        "change_master_copy",
        "add_delegate",
        "remove_delegate",
        "send_ether",
        "send_erc20",
        "send_erc721",
        "remove_proposed_transaction",
        "fs_attestation_service_auth",
        "fs_attestation_service_enrolment",
        "fs_attestation_service_deploy_guard_contract",
        "fs_attestation_service_set_constraints",
        "fs_attestation_settings"
    }

    def get_tokens_unprocessed(self, text: str) -> (int, Token, str):
        for index, token, value in BashLexer.get_tokens_unprocessed(self, text):
            if token is Text and value in self.EXTRA_KEYWORDS:
                yield index, Name.Builtin, value
            elif token is Text and re.search(self.ADDRESS, value):
                yield index, Keyword, value
            else:
                yield index, token, value



import functools

from trezorlib.exceptions import (
    Cancelled,
    OutdatedFirmwareError,
    PinException,
    TrezorFailure,
)
from trezorlib.transport import TransportException

from ..exceptions import HardwareWalletException
from .exceptions import InvalidDerivationPath


def raise_trezor_exception_as_hw_wallet_exception(function):
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except TrezorFailure as e:
            raise HardwareWalletException(e.message)
        except OutdatedFirmwareError:
            raise HardwareWalletException("Trezor firmware version is not supported")
        except PinException:
            raise HardwareWalletException("Wrong PIN")
        except Cancelled:
            raise HardwareWalletException("Trezor operation was cancelled")
        except TransportException:
            raise HardwareWalletException("Trezor device is not connected")
        except InvalidDerivationPath as e:
            raise HardwareWalletException(e.message)

    return wrapper



import os
import unittest
from unittest import mock
from unittest.mock import MagicMock

from eth_account import Account
from hexbytes import HexBytes
from ledgerblue.Dongle import Dongle
from ledgereth.exceptions import (
    LedgerAppNotOpened,
    LedgerCancel,
    LedgerLocked,
    LedgerNotFound,
)
from ledgereth.objects import LedgerAccount, SignedTransaction

from gnosis.eth.eip712 import eip712_encode
from gnosis.safe import SafeTx
from gnosis.safe.signatures import signature_split
from gnosis.safe.tests.safe_test_case import SafeTestCaseMixin

from safe_cli.operators.exceptions import HardwareWalletException
from safe_cli.operators.hw_wallets.ledger_wallet import LedgerWallet


class Testledger_wallet(SafeTestCaseMixin, unittest.TestCase):
    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.init_dongle",
        return_value=Dongle(),
    )
    def test_setup_ledger_wallet(self, mock_init_dongle: MagicMock):
        derivation_path = "44'/60'/0'/0"
        address = Account.create().address
        with self.assertRaises(HardwareWalletException):
            LedgerWallet(derivation_path)
        with mock.patch(
            "safe_cli.operators.hw_wallets.ledger_wallet.get_account_by_path",
            return_value=LedgerAccount(derivation_path, address),
        ):
            ledger_wallet = LedgerWallet(derivation_path)
            self.assertIsNotNone(ledger_wallet.dongle)
            self.assertEqual(ledger_wallet.address, address)

    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.sign_typed_data_draft",
        autospec=True,
    )
    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.get_account_by_path",
        autospec=True,
    )
    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.init_dongle",
        return_value=Dongle(),
    )
    def test_hw_device_exception(
        self,
        mock_init_dongle: MagicMock,
        mock_get_account_by_path: MagicMock,
        mock_sign: MagicMock,
    ):
        derivation_path = "44'/60'/0'/0"
        address = Account.create().address
        random_domain_bytes = os.urandom(32)
        random_message_bytes = os.urandom(32)
        mock_get_account_by_path.side_effect = LedgerNotFound
        with self.assertRaises(HardwareWalletException):
            LedgerWallet(derivation_path)

        mock_get_account_by_path.side_effect = LedgerLocked
        with self.assertRaises(HardwareWalletException):
            LedgerWallet(derivation_path)

        mock_get_account_by_path.side_effect = LedgerAppNotOpened
        with self.assertRaises(HardwareWalletException):
            LedgerWallet(derivation_path)

        # Test sign exceptions
        mock_get_account_by_path.side_effect = None
        mock_get_account_by_path.return_value = LedgerAccount(derivation_path, address)
        mock_sign.side_effect = LedgerNotFound
        with self.assertRaises(HardwareWalletException):
            ledger_wallet = LedgerWallet(derivation_path)
            ledger_wallet.sign_typed_hash(random_domain_bytes, random_message_bytes)

        mock_sign.side_effect = LedgerLocked
        with self.assertRaises(HardwareWalletException):
            ledger_wallet = LedgerWallet(derivation_path)
            ledger_wallet.sign_typed_hash(random_domain_bytes, random_message_bytes)

        mock_sign.side_effect = LedgerAppNotOpened
        with self.assertRaises(HardwareWalletException):
            ledger_wallet = LedgerWallet(derivation_path)
            ledger_wallet.sign_typed_hash(random_domain_bytes, random_message_bytes)

        mock_sign.side_effect = LedgerCancel
        with self.assertRaises(HardwareWalletException):
            ledger_wallet = LedgerWallet(derivation_path)
            ledger_wallet.sign_typed_hash(random_domain_bytes, random_message_bytes)

    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.get_account_by_path",
        autospec=True,
    )
    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.init_dongle",
        autospec=True,
        return_value=Dongle(),
    )
    def test_sign_typed_hash(
        self, mock_init_dongle: MagicMock, mock_get_account_by_path: MagicMock
    ):
        owner = Account.create()
        to = Account.create()
        derivation_path = "44'/60'/0'/0"
        mock_get_account_by_path.return_value = LedgerAccount(
            derivation_path, owner.address
        )
        ledger_wallet = LedgerWallet(derivation_path)

        safe = self.deploy_test_safe(
            owners=[owner.address],
            threshold=1,
            initial_funding_wei=self.w3.to_wei(0.1, "ether"),
        )
        safe_tx = SafeTx(
            self.ethereum_client,
            safe.address,
            to.address,
            10,
            b"",
            0,
            200000,
            200000,
            self.gas_price,
            None,
            None,
            safe_nonce=0,
        )
        encode_hash = eip712_encode(safe_tx.eip712_structured_data)
        expected_signature = safe_tx.sign(owner.key)
        # We need to split to change the bytes signature order to v + r + s like ledger return signature
        v, r, s = signature_split(expected_signature)

        ledger_return_signature = (
            v.to_bytes(1, byteorder="big")
            + r.to_bytes(32, byteorder="big")
            + s.to_bytes(32, byteorder="big")
        )
        mock_init_dongle.return_value.exchange = MagicMock(
            return_value=ledger_return_signature
        )
        signature = ledger_wallet.sign_typed_hash(encode_hash[1], encode_hash[2])
        self.assertEqual(expected_signature, signature)

        # Check that dongle exchange is called with the expected payload
        # https://github.com/LedgerHQ/app-ethereum/blob/master/doc/ethapp.adoc#sign-eth-eip-712
        command = "e00c0000" + "51"  # command + payload length
        payload = (
            "04" + "8000002c8000003c8000000000000000"
        )  # number of derivations + 44'/60'/0'/0
        expected_exchange_payload = (
            bytes.fromhex(command)
            + bytes.fromhex(payload)
            + encode_hash[1]
            + encode_hash[2]
        )
        mock_init_dongle.return_value.exchange.assert_called_once_with(
            expected_exchange_payload
        )

    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.create_transaction",
        autospec=True,
        return_value=Dongle(),
    )
    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.get_account_by_path",
        autospec=True,
    )
    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.init_dongle",
        autospec=True,
        return_value=Dongle(),
    )
    def test_get_signed_raw_transaction(
        self,
        mock_init_dongle: MagicMock,
        mock_get_account_by_path: MagicMock,
        mock_ledger_create_transaction: MagicMock,
    ):
        owner = Account.create()
        to = Account.create()
        derivation_path = "44'/60'/0'/0"
        mock_get_account_by_path.return_value = LedgerAccount(
            derivation_path, owner.address
        )
        ledger_wallet = LedgerWallet(derivation_path)
        safe = self.deploy_test_safe(
            owners=[owner.address],
            threshold=1,
            initial_funding_wei=self.w3.to_wei(0.1, "ether"),
        )
        safe_tx = SafeTx(
            self.ethereum_client,
            safe.address,
            to.address,
            10,
            b"",
            0,
            200000,
            200000,
            self.gas_price,
            None,
            None,
            safe_nonce=0,
        )
        safe_tx.sign(owner.key)
        tx_parameters = {
            "from": owner.address,
            "gasPrice": safe_tx.w3.eth.gas_price,
            "nonce": 0,
            "gas": safe_tx.recommended_gas(),
        }
        safe_tx.tx = safe_tx.w3_tx.build_transaction(tx_parameters)
        signed_fields = safe_tx.w3.eth.account.sign_transaction(
            safe_tx.tx, private_key=owner.key
        )

        mocked_signed_transaction_response = SignedTransaction(
            nonce=0,
            gas_price=safe_tx.tx["gasPrice"],
            gas_limit=safe_tx.tx["gas"],
            destination=HexBytes(safe_tx.tx["to"]),
            amount=safe_tx.tx["value"],
            data=HexBytes(safe_tx.tx["data"]),
            v=signed_fields.v,
            r=signed_fields.r,
            s=signed_fields.s,
        )
        mock_ledger_create_transaction.return_value = mocked_signed_transaction_response

        raw_signed_tx = ledger_wallet.get_signed_raw_transaction(
            safe_tx.tx, safe_tx.ethereum_client.get_chain_id()
        )  # return raw signed transaction
        self.assertEqual(signed_fields.rawTransaction, HexBytes(raw_signed_tx))



#!/bin/env python3
import argparse
import secrets
import sys
from typing import List

from art import text2art
from eth_account import Account
from eth_account.signers.local import LocalAccount
from eth_typing import URI
from hexbytes import HexBytes
from prompt_toolkit import print_formatted_text

from gnosis.eth import EthereumClient, EthereumTxSent
from gnosis.eth.constants import NULL_ADDRESS
from gnosis.eth.contracts import get_safe_V1_4_1_contract
from gnosis.safe import ProxyFactory, Safe

from safe_cli.safe_addresses import (
    get_default_fallback_handler_address,
    get_proxy_factory_address,
    get_safe_contract_address,
    get_safe_l2_contract_address,
)
from safe_cli.utils import yes_or_no_question

from .argparse_validators import (
    check_ethereum_address,
    check_positive_integer,
    check_private_key,
)


def setup_argument_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("node_url", help="Ethereum node url")
    parser.add_argument(
        "private_key", help="Deployer private_key", type=check_private_key
    )
    parser.add_argument(
        "--threshold",
        help="Number of owners required to execute transactions on the created Safe. It must"
        "be greater than 0 and less or equal than the number of owners",
        type=check_positive_integer,
        default=1,
    )
    parser.add_argument(
        "--owners",
        help="Owners. By default it will be just the deployer",
        nargs="+",
        type=check_ethereum_address,
    )
    parser.add_argument(
        "--safe-contract",
        help="Use a custom Safe master copy",
        default=None,
        type=check_ethereum_address,
    )
    parser.add_argument(
        "--proxy-factory",
        help="Use a custom proxy factory",
        default=None,
        type=check_ethereum_address,
    )
    parser.add_argument(
        "--callback-handler",
        help="Use a custom fallback handler. It is not required for Safe Master Copies "
        "with version < 1.1.0",
        default=None,
        type=check_ethereum_address,
    )
    parser.add_argument(
        "--salt-nonce",
        help="Use a custom nonce for the deployment. Same nonce with same deployment configuration will "
        "lead to the same Safe address ",
        default=secrets.randbits(256),
        type=int,
    )

    parser.add_argument(
        "--without-events",
        help="Use non events deployment of the Safe instead of the regular one. Recommended for mainnet to save gas costs when using the Safe",
        default=False,
        action="store_true",
    )
    return parser


def main(*args, **kwargs) -> EthereumTxSent:
    parser = setup_argument_parser()
    print_formatted_text(text2art("Safe Creator"))  # Print fancy text
    args = parser.parse_args()
    node_url: URI = args.node_url
    account: LocalAccount = Account.from_key(args.private_key)
    owners: List[str] = args.owners if args.owners else [account.address]
    threshold: int = args.threshold
    salt_nonce: int = args.salt_nonce
    to = NULL_ADDRESS
    data = b""
    payment_token = NULL_ADDRESS
    payment = 0
    payment_receiver = NULL_ADDRESS

    if len(owners) < threshold:
        print_formatted_text(
            "Threshold cannot be bigger than the number of unique owners"
        )
        sys.exit(1)

    ethereum_client = EthereumClient(node_url)
    ethereum_network = ethereum_client.get_network()

    safe_contract_address = args.safe_contract or (
        get_safe_contract_address(ethereum_client)
        if args.without_events
        else get_safe_l2_contract_address(ethereum_client)
    )
    proxy_factory_address = args.proxy_factory or get_proxy_factory_address(
        ethereum_client
    )
    fallback_handler = args.callback_handler or get_default_fallback_handler_address(
        ethereum_client
    )

    if not ethereum_client.is_contract(safe_contract_address):
        print_formatted_text(
            f"Safe contract address {safe_contract_address} "
            f"does not exist on network {ethereum_network.name}"
        )
        sys.exit(1)
    elif not ethereum_client.is_contract(proxy_factory_address):
        print_formatted_text(
            f"Proxy contract address {proxy_factory_address} "
            f"does not exist on network {ethereum_network.name}"
        )
        sys.exit(1)
    elif fallback_handler != NULL_ADDRESS and not ethereum_client.is_contract(
        fallback_handler
    ):
        print_formatted_text(
            f"Fallback handler address {fallback_handler} "
            f"does not exist on network {ethereum_network.name}"
        )
        sys.exit(1)

    account_balance: int = ethereum_client.get_balance(account.address)
    if not account_balance:
        print_formatted_text(
            "Client does not have any funds. Let's try anyway in case it's a network without gas costs"
        )
    else:
        ether_account_balance = round(
            ethereum_client.w3.from_wei(account_balance, "ether"), 6
        )
        print_formatted_text(
            f"Network {ethereum_client.get_network().name} - Sender {account.address} - "
            f"Balance: {ether_account_balance}Ξ"
        )

    if not ethereum_client.w3.eth.get_code(
        safe_contract_address
    ) or not ethereum_client.w3.eth.get_code(proxy_factory_address):
        print_formatted_text("Network not supported")
        sys.exit(1)

    print_formatted_text(
        f"Creating new Safe with owners={owners} threshold={threshold} salt-nonce={salt_nonce}"
    )
    safe_version = Safe(safe_contract_address, ethereum_client).retrieve_version()
    print_formatted_text(
        f"Safe-master-copy={safe_contract_address} version={safe_version}\n"
        f"Fallback-handler={fallback_handler}\n"
        f"Proxy factory={proxy_factory_address}"
    )
    if yes_or_no_question("Do you want to continue?"):
        safe_contract = get_safe_V1_4_1_contract(
            ethereum_client.w3, safe_contract_address
        )
        safe_creation_tx_data = HexBytes(
            safe_contract.functions.setup(
                owners,
                threshold,
                to,
                data,
                fallback_handler,
                payment_token,
                payment,
                payment_receiver,
            ).build_transaction({"gas": 1, "gasPrice": 1})["data"]
        )

        proxy_factory = ProxyFactory(proxy_factory_address, ethereum_client)
        expected_safe_address = proxy_factory.calculate_proxy_address(
            safe_contract_address, safe_creation_tx_data, salt_nonce
        )
        if ethereum_client.is_contract(expected_safe_address):
            print_formatted_text(f"Safe on {expected_safe_address} is already deployed")
            sys.exit(1)

        if yes_or_no_question(
            f"Safe will be deployed on {expected_safe_address}, looks good?"
        ):
            ethereum_tx_sent = proxy_factory.deploy_proxy_contract_with_nonce(
                account, safe_contract_address, safe_creation_tx_data, salt_nonce
            )
            print_formatted_text(
                f"Sent tx with tx-hash={ethereum_tx_sent.tx_hash.hex()} "
                f"Safe={ethereum_tx_sent.contract_address} is being created"
            )
            print_formatted_text(f"Tx parameters={ethereum_tx_sent.tx}")
            return ethereum_tx_sent



import functools

from ledgereth.exceptions import (
    LedgerAppNotOpened,
    LedgerCancel,
    LedgerLocked,
    LedgerNotFound,
)

from ..exceptions import HardwareWalletException
from .exceptions import InvalidDerivationPath


def raise_ledger_exception_as_hw_wallet_exception(function):
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except LedgerNotFound as e:
            raise HardwareWalletException(e.message)
        except LedgerLocked as e:
            raise HardwareWalletException(e.message)
        except LedgerAppNotOpened as e:
            raise HardwareWalletException(e.message)
        except LedgerCancel as e:
            raise HardwareWalletException(e.message)
        except InvalidDerivationPath as e:
            raise HardwareWalletException(e.message)
        except BaseException as e:
            if "Error while writing" in e.args:
                raise HardwareWalletException("Ledger error writting, restart safe-cli")
            raise e

    return wrapper
